# ai_hydra/utils/HydraBaseMQ.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0
#

from typing import Any
from collections.abc import Awaitable, Callable

import time
import asyncio
import zmq
import zmq.asyncio
from zmq.sugar.frame import Frame

from ai_hydra.constants.DHydra import (
    DHydraRouterDef,
    DModule,
    DMethod,
    DHydra,
)
from ai_hydra.constants.DHydraMQ import DHydraMQ, DHydraMQDef
from ai_hydra.zmq.HydraMsg import HydraMsg

MsgHandler = Callable[[HydraMsg], Any | Awaitable[Any]]


class HydraBaseMQ:
    def __init__(
        self,
        router_address: str = DHydraRouterDef.HOSTNAME,
        router_port: int = DHydraRouterDef.PORT,
        router_hb_port: int = DHydraRouterDef.HEARTBEAT_PORT,
        identity: str = DModule.HYDRA_MQ,
        topic_prefix: str = DHydraMQDef.TOPIC_PREFIX,
    ) -> None:

        self.topic_prefix = topic_prefix
        self.router = router_address
        self.port = router_port
        self.hb_port = router_hb_port
        self.identity = identity

        self.router_addr = f"tcp://{self.router}:{self.port}"
        self.router_hb_addr = f"tcp://{self.router}:{self.hb_port}"

        self.ctx = zmq.asyncio.Context()
        self.socket = self.ctx.socket(zmq.DEALER)
        self.hb_socket = self.ctx.socket(zmq.DEALER)

        self.socket.setsockopt(
            zmq.IDENTITY, self.identity.encode(DHydraMQ.UTF_8)
        )
        self.hb_socket.setsockopt(
            zmq.IDENTITY, self.identity.encode(DHydraMQ.UTF_8)
        )

        self.socket.connect(self.router_addr)
        self.hb_socket.connect(self.router_hb_addr)

        self.hb_task: asyncio.Task[None] | None = None
        self.hb_stop_event = asyncio.Event()
        self._last_heartbeat = 0.0
        self._started = False
        self._stopped = False

    @staticmethod
    def _ensure_bytes(data: bytes | Frame) -> bytes:
        return data.bytes if isinstance(data, Frame) else data

    @staticmethod
    def _ignore_zmq_teardown(action: Callable[[], None], what: str) -> None:
        try:
            action()
        except zmq.ZMQError as e:
            # expected during shutdown races / already-closed sockets
            print(
                f"DEBUG: ignoring {what} during shutdown: {type(e).__name__}: {e}"
            )

    def connected(self) -> bool:
        return (time.time() - self._last_heartbeat) < (
            2 * DHydra.HEARTBEAT_INTERVAL
        )

    async def quit(self) -> None:
        if self._stopped:
            return
        self._stopped = True
        self._started = False
        self.hb_stop_event.set()

        if self.hb_task is not None:
            self.hb_task.cancel()
            try:
                await asyncio.wait_for(self.hb_task, timeout=1.0)
            except asyncio.TimeoutError:
                print("DEBUG: hb_task did not cancel cleanly")
            except asyncio.CancelledError:
                pass
            except Exception as e:
                print(
                    f"DEBUG: hb_task exception during quit: {type(e).__name__}: {e}"
                )
            finally:
                self.hb_task = None

        self._ignore_zmq_teardown(
            lambda: self.hb_socket.disconnect(self.router_hb_addr),
            f"hb_socket.disconnect({self.router_hb_addr})",
        )
        self._ignore_zmq_teardown(
            lambda: self.hb_socket.close(linger=0),
            "hb_socket.close(linger=0)",
        )

        self._ignore_zmq_teardown(
            lambda: self.socket.disconnect(self.router_addr),
            f"socket.disconnect({self.router_addr})",
        )
        self._ignore_zmq_teardown(
            lambda: self.socket.close(linger=0),
            "socket.close(linger=0)",
        )

    async def recv(self) -> HydraMsg:
        message_data = None
        message_data = await asyncio.wait_for(
            self.socket.recv(copy=True), timeout=DHydra.NETWORK_TIMEOUT
        )
        if message_data is not None:
            return HydraMsg.from_json(self._ensure_bytes(message_data))

    async def send(self, msg: HydraMsg) -> None:
        await self.socket.send(msg.to_json())

    def start(self) -> None:
        if self._stopped:
            raise RuntimeError("HydraBaseMQ cannot be restarted after quit()")

        if self._started:
            return

        self.hb_task = asyncio.create_task(
            self.start_hb_bg(), name=DHydraMQ.HEARTBEAT
        )
        self._started = True

    async def start_hb_bg(self) -> None:
        try:
            while not self.hb_stop_event.is_set():
                msg = HydraMsg(
                    sender=self.identity,
                    target=DModule.HYDRA_ROUTER,
                    method=DMethod.HEARTBEAT,
                )
                await self.hb_socket.send(msg.to_json())

                try:
                    message_data = await asyncio.wait_for(
                        self.hb_socket.recv(copy=True),
                        timeout=DHydra.NETWORK_TIMEOUT,
                    )
                    reply = HydraMsg.from_json(
                        self._ensure_bytes(message_data)
                    )

                    if reply.method == DMethod.HEARTBEAT_REPLY:
                        self._last_heartbeat = time.time()

                except asyncio.CancelledError:
                    raise
                except asyncio.TimeoutError:
                    pass

                await asyncio.sleep(DHydra.HEARTBEAT_INTERVAL)

        except asyncio.CancelledError:
            raise

    def topic(self, suffix: str) -> str:
        return f"{self.topic_prefix}.{suffix}"



# ai_hydra/client/HydraClient.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

# General Python Modules
import asyncio
import sys
import time
from pathlib import Path
import os
from datetime import datetime

# Textual Modules
from textual import work, on
from textual.app import App, ComposeResult
from textual.containers import Horizontal, Vertical
from textual.theme import Theme
from textual.widgets import (
    Button,
    Label,
    Input,
    Checkbox,
    Select,
    Switch,
)
from textual.message import Message

# AI Hydra Constants
from ai_hydra.constants.DHydra import (
    DHydra,
    DHydraRouterDef,
    DMethod,
    DModule,
)
from ai_hydra.constants.DHydraMQ import DHydraMQDef, DHydraMQ, DEvent
from ai_hydra.constants.DHydraTui import DField, DFile, DLabel, DStatus
from ai_hydra.constants.DGame import DGameField, DGameMethod
from ai_hydra.constants.DNNet import (
    DNetField,
    DNetDef,
    DLinear,
    DRNN,
    MODEL_TYPE_TABLE,
    MODEL_TYPES,
)

# AI Hydra Modules
from ai_hydra.zmq.HydraClientMQ import HydraClientMQ
from ai_hydra.zmq.HydraMsg import HydraMsg
from ai_hydra.utils.SimCfg import SimCfg
from ai_hydra.utils.HydraSnapshot import HydraSnapshot
from ai_hydra.utils.HydraMetrics import HydraMetrics
from ai_hydra.client.ClientGameBoard import ClientGameBoard
from ai_hydra.client.HighScoresLog import HighScoresLog
from ai_hydra.client.HydraTelemetry import HydraTelemetry


HYDRA_THEME = Theme(
    name="hydra_theme",
    primary="#0E191C",
    secondary="#2aa5ceff",
    accent="#B48EAD",
    foreground="#31b8e6",
    background="black",
    success="#A3BE8C",
    warning="#EBCB8B",
    error="#BF616A",
    surface="#2aa5ce",
    panel="#000000",
    dark=True,
    variables={
        "block-cursor-text-style": "none",
        "footer-key-foreground": "#FF0000",
        "input-selection-background": "#81a1c1 35%",
    },
)


class TelemetryReceived(Message):
    def __init__(self, topic: str, payload: dict) -> None:
        super().__init__()
        self.topic = topic
        self.payload = payload


class HydraClientTui(App):
    """A Textual interface to the HydraServer"""

    TITLE = DLabel.CLIENT_TITLE
    CSS_PATH = DFile.CLIENT_CSS

    def __init__(
        self,
        address: str = DHydraRouterDef.HOSTNAME,
        port: int = DHydraRouterDef.PORT,
    ) -> None:
        """Constructor"""
        super().__init__()

        self._address: str = address
        self._port = port
        self._identity = DModule.HYDRA_CLIENT
        self._connected_msg = DStatus.BAD + " " + DLabel.DISCONNECTED
        self.mq: HydraClientMQ | None = None

        self.game_board = ClientGameBoard(20, id=DGameField.BOARD)
        self.cfg = SimCfg()
        self._running = False
        self._wgt = None

        # Metrics to telemetry data
        self.metrics = HydraMetrics(initial_epsilon=DLinear.INITIAL_EPSILON)

        # Snapshot class to create a snapshot file.
        self.snapshot = HydraSnapshot(metrics=self.metrics)

        # Telemetric Widget
        self.telemtry: HydraTelemetry | None = None

        self._cur_epsilon = None

        # ZeroMQ batching support at this app layer it checks for lost
        # messages. PUB/SUB does not guarantee delivery, batching  aims to
        # mitigate that.
        self._prev_scores_batch_num = None
        self._cur_scores_batch_num = None
        self._first_scores_batch = True

        self._prev_per_ep_batch_num = None
        self._cur_per_ep_batch_num = None
        self._first_per_ep_batch = True

        # AI Hydra stores temporary and persistent files in this directory
        self._hydra_dir = os.path.join(Path.home(), DHydra.HYDRA_DIR)
        if not os.path.exists(self._hydra_dir):
            os.mkdir(self._hydra_dir)

    @work(exclusive=True)
    async def check_heartbeat(self) -> None:
        while True:
            if self.mq.connected():
                status = DStatus.GOOD
            else:
                status = DStatus.BAD

            self.query_one(f"#{DField.ROUTER_HB}", Label).update(
                f"{DLabel.ROUTER}: {status}"
            )

            await asyncio.sleep(DHydra.HEARTBEAT_INTERVAL)

    def compose(self) -> ComposeResult:
        """The TUI is created here"""

        ## We're using a Textual "grid" layout...

        # ----- Title ---
        yield Label(DLabel.CLIENT_TITLE, id=DField.TITLE)

        # ----- Buttons ---
        yield Vertical(
            Button(label=DLabel.HANDSHAKE, id=DField.HANDSHAKE, compact=True),
            Button(label=DLabel.START, id=DField.START_RUN, compact=True),
            Label(id=DField.START_RUN_DIVIDER),
            Button(label=DLabel.SNAPSHOT, id=DField.SNAPSHOT, compact=True),
            Label(),
            Button(label=DLabel.QUIT, id=DMethod.QUIT, compact=True),
            id=DField.BUTTONS,
        )

        # ------ The Snake Game ---
        yield Vertical(self.game_board, id=DField.BOARD_BOX)

        # ----- Settings ---
        yield Vertical(
            # Random seed
            Horizontal(
                Label(f"{DLabel.RANDOM_SEED:>11s}: "),
                Input(
                    type=DField.INTEGER,
                    compact=True,
                    valid_empty=False,
                    value=str(DHydra.RANDOM_SEED),
                    id=DField.RANDOM_SEED_INPUT,
                ),
                Label(str(DHydra.RANDOM_SEED), id=DField.RANDOM_SEED_LABEL),
                classes=DField.INPUT_FIELD,
            ),
            # Move delay
            Horizontal(
                Label(f"{DLabel.MOVE_DELAY:>11s}: "),
                Label(f"0.0", id=DField.MOVE_DELAY_LABEL),
                Input(
                    type=DField.NUMBER,
                    compact=True,
                    valid_empty=False,
                    value=str(DNetDef.MOVE_DELAY),
                    id=DField.MOVE_DELAY_INPUT,
                ),
                classes=DField.INPUT_FIELD,
            ),
            # Turbo mode
            Horizontal(
                Label(f"{DLabel.TURBO_MODE:>11s}: "),
                # Checkbox(value=False, id=DField.TURBO_MODE, compact=True),
                Switch(value=False, id=DField.TURBO_MODE),
                classes=DField.INPUT_FIELD,
            ),
            id=DField.SETTINGS,
        )

        # ----- Highscores widget ---
        yield HighScoresLog(id=DField.HIGHSCORES_LOG)

        # ----- Epsilon ---
        yield Vertical(
            # Initial epsilon
            Horizontal(
                Label(f"{DLabel.INITIAL_EPSILON:>15s}: "),
                Label(
                    str(DLinear.INITIAL_EPSILON),
                    id=DField.INITIAL_EPSILON_LABEL,
                ),
                Input(
                    type=DField.NUMBER,
                    compact=True,
                    valid_empty=False,
                    value=str(DLinear.INITIAL_EPSILON),
                    id=DField.INITIAL_EPSILON_INPUT,
                ),
                classes=DField.INPUT_FIELD,
            ),
            # Minimum epsilon
            Horizontal(
                Label(f"{DLabel.MIN_EPSILON:>15s}: "),
                Label(
                    str(DLinear.MINIMUM_EPSILON),
                    id=DField.MIN_EPSILON_LABEL,
                ),
                Input(
                    type=DField.NUMBER,
                    compact=True,
                    valid_empty=False,
                    value=str(DLinear.MINIMUM_EPSILON),
                    id=DField.MIN_EPSILON_INPUT,
                ),
                classes=DField.INPUT_FIELD,
            ),
            # Epsilon decay
            Horizontal(
                Label(f"{DLabel.EPSILON_DECAY:>15s}: "),
                Label(
                    str(DLinear.EPSILON_DECAY_RATE),
                    id=DField.EPSILON_DECAY_LABEL,
                ),
                Input(
                    type=DField.NUMBER,
                    compact=True,
                    valid_empty=False,
                    value=str(DLinear.EPSILON_DECAY_RATE),
                    id=DField.EPSILON_DECAY_INPUT,
                ),
                classes=DField.INPUT_FIELD,
            ),
            # Current epsilon
            Horizontal(
                Label(f"{DLabel.CUR_EPSILON:>15s}: "),
                Label(id=DField.CUR_EPSILON),
                classes=DField.INPUT_FIELD,
            ),
            id=DField.EPSILON,
        )

        # ----- Model ---
        yield Vertical(
            # Model type
            Horizontal(
                Label(f"{DLabel.NN_MODEL:>15s}: "),
                Label(DLabel.LINEAR, id=DField.MODEL_TYPE_LABEL),
                Select(
                    MODEL_TYPES,
                    compact=True,
                    id=DField.MODEL_TYPE_SELECT,
                    allow_blank=False,
                ),
                classes=DField.INPUT_FIELD,
            ),
            # Hidden Size
            Horizontal(
                Label(f"{DLabel.HIDDEN_SIZE:>15s}: "),
                Label(
                    f"{DLinear.HIDDEN_SIZE:.7f}",
                    id=DField.HIDDEN_SIZE_LABEL,
                ),
                Input(
                    type=DField.INTEGER,
                    compact=True,
                    valid_empty=False,
                    value=f"{DLinear.HIDDEN_SIZE}",
                    id=DField.HIDDEN_SIZE_INPUT,
                ),
                classes=DField.INPUT_FIELD,
            ),
            # Dropout p-value
            Horizontal(
                Label(f"{DLabel.DROPOUT_P_VAL:>15s}: "),
                Label(
                    f"{DLinear.DROPOUT_P:.2f}",
                    id=DField.DROPOUT_P_LABEL,
                ),
                Input(
                    type=DField.NUMBER,
                    compact=True,
                    valid_empty=False,
                    value=f"{DLinear.DROPOUT_P:.2f}",
                    id=DField.DROPOUT_P_INPUT,
                ),
                classes=DField.INPUT_FIELD,
            ),
            # RNN Layers
            Horizontal(
                Label(f"{DLabel.RNN_LAYERS:>15s}: ", id=DField.RNN_LAYERS_OPT),
                Label(
                    f"{DRNN.RNN_LAYERS}",
                    id=DField.RNN_LAYERS_LABEL,
                ),
                Input(
                    type=DField.INTEGER,
                    compact=True,
                    valid_empty=False,
                    value=f"{DRNN.RNN_LAYERS}",
                    id=DField.RNN_LAYERS_INPUT,
                ),
                classes=DField.INPUT_FIELD,
            ),
            id=DField.MODEL,
        )

        # ------ Network ---
        yield Vertical(
            Label(f"{DLabel.TARGET_HOST}: {self._address}"),
            Label(f"{DLabel.TARGET_PORT}: {self._port}"),
            Label(f"{DLabel.ROUTER}", id=DField.ROUTER_HB),
            id=DField.NETWORK,
        )

        # ----- Training ---
        yield Vertical(
            # Learning rate
            Horizontal(
                Label(f"{DLabel.LEARNING_RATE:>15s}: "),
                Label(
                    f"{DLinear.LEARNING_RATE:.7f}",
                    id=DField.LEARNING_RATE_LABEL,
                ),
                Input(
                    type=DField.NUMBER,
                    compact=True,
                    valid_empty=False,
                    value=f"{DLinear.LEARNING_RATE:.7f}",
                    id=DField.LEARNING_RATE_INPUT,
                ),
                classes=DField.INPUT_FIELD,
            ),
            # Discount/Gamma
            Horizontal(
                Label(f"{DLabel.GAMMA:>15s}: "),
                Label(
                    f"{DLinear.GAMMA:.2f}",
                    id=DField.GAMMA_LABEL,
                ),
                Input(
                    type=DField.NUMBER,
                    compact=True,
                    valid_empty=False,
                    value=f"{DLinear.GAMMA:.2f}",
                    id=DField.GAMMA_INPUT,
                ),
                classes=DField.INPUT_FIELD,
            ),
            # Batch size
            Horizontal(
                Label(f"{DLabel.BATCH_SIZE:>15s}: "),
                Label(
                    f"{DLinear.BATCH_SIZE}",
                    id=DField.BATCH_SIZE_LABEL,
                ),
                Input(
                    type=DField.INTEGER,
                    compact=True,
                    valid_empty=False,
                    value=f"{DLinear.BATCH_SIZE}",
                    id=DField.BATCH_SIZE_INPUT,
                ),
                classes=DField.INPUT_FIELD,
            ),
            # RNN Sequence Length
            Horizontal(
                Label(
                    f"{DLabel.SEQUENCE_LENGTH:>15s}: ",
                    id=DField.SEQ_LENGTH_OPT,
                ),
                Label(
                    f"{DRNN.SEQ_LENGTH}",
                    id=DField.SEQ_LENGTH_LABEL,
                ),
                Input(
                    type=DField.NUMBER,
                    compact=True,
                    valid_empty=False,
                    value=f"{DRNN.SEQ_LENGTH}",
                    id=DField.SEQ_LENGTH_INPUT,
                ),
                classes=DField.INPUT_FIELD,
            ),
            # RNN Trainer Tau
            Horizontal(
                Label(f"{DLabel.RNN_TAU:>15s}: ", id=DField.RNN_TAU_OPT),
                Label(
                    f"{DRNN.TAU}",
                    id=DField.RNN_TAU_LABEL,
                ),
                Input(
                    type=DField.NUMBER,
                    compact=True,
                    valid_empty=False,
                    value=f"{DRNN.TAU}",
                    id=DField.RNN_TAU_INPUT,
                ),
                classes=DField.INPUT_FIELD,
            ),
            id=DField.TRAINING_BOX,
        )

        # Realtime Telemetry (plots and event log)
        yield HydraTelemetry(metrics=self.metrics, id=DField.HYDRA_TELEMETRY)

        # Focus widget: This is hidden, but it allows me to move focus away
        # from the selected button when a button is clicked.
        yield Checkbox(id=DField.HIDDEN_WIDGET)

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        button_id = event.button.id

        if self.mq is None:
            raise TypeError("self.mq is None!!!")

        elif button_id == DField.HANDSHAKE:
            await self._send_handshake()

        elif button_id == DMethod.QUIT:
            await self.on_quit()

        elif button_id == DGameMethod.RESET_GAME:
            await self._send_reset()

        elif button_id == DField.SNAPSHOT:
            await self._take_snapshot()

        elif button_id == DGameMethod.START_RUN:
            self._update_tui_labels()
            self.mq.enable_per_episode_sub()
            self.mq.enable_scores_sub()
            self.mq.enable_events_sub()
            if self.cfg.get(DNetField.PER_STEP):
                self.mq.enable_per_step_sub()
            else:
                self.mq.disable_per_step_sub()
            await self._send_start_run()

        elif button_id == DGameMethod.STOP_RUN:
            await self._send_stop_run()

        elif button_id == DField.UPDATE_RUNTIME_CONFIG:
            self._update_tui_labels()
            await self._send_update_config()
        self._w_hidden_widget.focus()

    async def on_input_changed(self, event: Input.Changed) -> None:
        input_id = event.input.id
        if input_id != DField.MOVE_DELAY_INPUT:
            return

        self._update_tui_labels()
        msg = HydraMsg(
            sender=DModule.HYDRA_CLIENT,
            target=DModule.HYDRA_MGR,
            method=DGameMethod.UPDATE_CONFIG,
            payload=self.cfg.to_dict(),
        )
        await self.mq.send(msg)

    def on_mount(self) -> None:
        self.add_class(DField.BAD_HANDSHAKE)
        self.add_class(DField.LINEAR)
        self.add_class(DField.TURBO_OFF)

        # Create references to TUI elements that are being updated
        self.telemtry = self.query_one(
            f"#{DField.HYDRA_TELEMETRY}", HydraTelemetry
        )
        self.event_log = self.telemtry.event_log
        self._w_batch_size_input = self.query_one(
            f"#{DField.BATCH_SIZE_INPUT}", Input
        )
        self._w_batch_size_label = self.query_one(
            f"#{DField.BATCH_SIZE_LABEL}", Label
        )
        self._w_board_box = self.query_one(f"#{DField.BOARD_BOX}", Vertical)
        self._w_cur_epsilon_label = self.query_one(
            f"#{DField.CUR_EPSILON}", Label
        )
        self._w_dropout_p_input = self.query_one(
            f"#{DField.DROPOUT_P_INPUT}", Input
        )
        self._w_dropout_p_label = self.query_one(
            f"#{DField.DROPOUT_P_LABEL}", Label
        )
        self._w_epsilon_decay_input = self.query_one(
            f"#{DField.EPSILON_DECAY_INPUT}", Input
        )
        self._w_epsilon_decay_label = self.query_one(
            f"#{DField.EPSILON_DECAY_LABEL}", Label
        )
        self._w_gamma_input = self.query_one(f"#{DField.GAMMA_INPUT}", Input)
        self._w_gamma_label = self.query_one(f"#{DField.GAMMA_LABEL}", Label)
        self._w_hidden_size_input = self.query_one(
            f"#{DField.HIDDEN_SIZE_INPUT}", Input
        )
        self._w_hidden_size_label = self.query_one(
            f"#{DField.HIDDEN_SIZE_LABEL}", Label
        )
        self._w_hidden_widget = self.query_one(
            f"#{DField.HIDDEN_WIDGET}", Checkbox
        )
        self._w_initial_epsilon_input = self.query_one(
            f"#{DField.INITIAL_EPSILON_INPUT}", Input
        )
        self._w_initial_epsilon_label = self.query_one(
            f"#{DField.INITIAL_EPSILON_LABEL}", Label
        )
        self._w_learning_rate_input = self.query_one(
            f"#{DField.LEARNING_RATE_INPUT}", Input
        )
        self._w_learning_rate_label = self.query_one(
            f"#{DField.LEARNING_RATE_LABEL}", Label
        )
        self._w_min_epsilon_input = self.query_one(
            f"#{DField.MIN_EPSILON_INPUT}", Input
        )
        self._w_min_epsilon_label = self.query_one(
            f"#{DField.MIN_EPSILON_LABEL}", Label
        )
        self._w_model_type_label = self.query_one(
            f"#{DField.MODEL_TYPE_LABEL}", Label
        )
        self._w_model_type_select = self.query_one(
            f"#{DField.MODEL_TYPE_SELECT}", Select
        )
        self._w_move_delay_input = self.query_one(
            f"#{DField.MOVE_DELAY_INPUT}", Input
        )
        self._w_random_seed_input = self.query_one(
            f"#{DField.RANDOM_SEED_INPUT}", Input
        )
        self._w_random_seed_label = self.query_one(
            f"#{DField.RANDOM_SEED_LABEL}", Label
        )
        self._w_rnn_layers_input = self.query_one(
            f"#{DField.RNN_LAYERS_INPUT}", Input
        )
        self._w_rnn_layers_label = self.query_one(
            f"#{DField.RNN_LAYERS_LABEL}", Label
        )
        self._w_rnn_tau_input = self.query_one(
            f"#{DField.RNN_TAU_INPUT}", Input
        )
        self._w_rnn_tau_label = self.query_one(
            f"#{DField.RNN_TAU_LABEL}", Label
        )
        self._w_seq_length_input = self.query_one(
            f"#{DField.SEQ_LENGTH_INPUT}", Input
        )
        self._w_seq_length_label = self.query_one(
            f"#{DField.SEQ_LENGTH_LABEL}", Label
        )
        self._w_highscores_log = self.query_one(
            f"#{DField.HIGHSCORES_LOG}", HighScoresLog
        )

        # Network
        self.query_one(f"#{DField.ROUTER_HB}", Label).update(
            f"{DLabel.ROUTER:>11s}: {DStatus.UNKNOWN}"
        )

        # Monitor the connection to the server in the background
        self.check_heartbeat()

        # ----- Add some text to the borders around the widgets
        self.query_one(f"#{DField.TITLE}").border_subtitle = (
            DLabel.VERSION + " " + DHydra.VERSION
        )
        self.query_one(f"#{DField.SETTINGS}").border_subtitle = DLabel.SETTINGS
        self.query_one(f"#{DField.NETWORK}").border_subtitle = DLabel.NETWORK
        self._w_highscores_log.border_subtitle = DLabel.HIGHSCORES
        self.query_one(f"#{DField.BUTTONS}").border_subtitle = DLabel.ACTIONS
        self.query_one(f"#{DField.MODEL}").border_subtitle = DLabel.MODEL
        self.query_one(f"#{DField.EPSILON}").border_subtitle = DLabel.EPSILON
        self.query_one(f"#{DField.TRAINING_BOX}").border_subtitle = (
            DLabel.TRAINING
        )
        # --------------------------------------------------------------

        # Switch focus to a hidden widget
        self._w_hidden_widget.focus()

        # Initialize ZeroMQ
        self.mq = HydraClientMQ(
            router_address=self._address,
            router_port=self._port,
            identity=self._identity,
            srv_host=self._address,
        )
        self.mq.sub_methods = {
            self.mq.topic(DHydraMQDef.EVENTS_TOPIC): self._on_sim_event,
            self.mq.topic(DHydraMQDef.PER_STEP_TOPIC): self.on_per_step,
            self.mq.topic(DHydraMQDef.PER_EPISODE_TOPIC): self.on_per_episode,
            self.mq.topic(DHydraMQDef.SCORES_TOPIC): self.on_scores,
        }
        self.mq.start()
        self.mq.disable_events_sub()
        self.mq.disable_per_episode_sub()
        self.mq.disable_per_step_sub()
        self.mq.disable_scores_sub()

        self.event_log.add_event(ev_type=DField.TUI, event="Initialized...")

    def on_per_episode(self, topic: str, payload: dict) -> None:
        # Unbatch messages
        batch_msg = HydraMsg.from_dict(payload)
        messages = batch_msg.payload[DHydraMQ.MSGS]
        self._cur_per_ep_batch_num = messages[0]["payload"]["count"]
        if self._first_per_ep_batch:
            self._first_per_ep_batch_num = False
            self._prev_per_ep_batch_num = self._cur_per_ep_batch_num - 1
        if (self._cur_per_ep_batch_num - self._prev_per_ep_batch_num) != 1:
            self.event_log.add_event(
                ev_type=DField.WARNING,
                event="Per episode data is being dropped",
            )
        self._prev_per_ep_batch_num = self._cur_per_ep_batch_num

        metrics = self.metrics

        # Process messages
        for payload in messages[1:]:

            # Get the data from the payload
            epoch = payload.get(DGameField.EPOCH)
            elapsed_time = payload.get(DGameField.ELAPSED_TIME)
            cur_epsilon = payload.get(DNetField.CUR_EPSILON)
            cur_loss = payload.get(DNetField.LOSS, None)

            # Load the data into the metrics object
            metrics.add_cur_epoch(epoch)
            metrics.add_elapsed_time(elapsed_time)
            metrics.add_cur_epsilon(cur_epsilon)
            is_new_data = metrics.add_cur_loss(cur_loss)
            if is_new_data:
                self.telemtry.loss_plot.plot_all()

            # Update the TUI
            self._w_board_box.border_title = f"{DLabel.GAME}: {epoch}"
            self.query_one(f"#{DField.HYDRA_TELEMETRY}").border_subtitle = (
                elapsed_time
            )
            self._w_cur_epsilon_label.update(str(round(cur_epsilon, 4)))

    def on_per_step(self, topic: str, payload: dict) -> None:
        # These messages are not batched (there's a MOVE_DELAY)
        board = payload.get(DGameField.BOARD)
        self.game_board.apply_board_dict(board)

    async def on_quit(self) -> None:
        ###
        ### The code below (clean shutdown) causes the client to hang
        ### requiring a `kill <PID>``, followed by a `tput reset` to get
        ### your terminal back into a sane state. So....
        ###
        ### "If in doubt, nuke it from orbit"
        ###
        if self.mq is not None:
            await self.mq.quit()
            self.mq = None
        self.exit()

    def on_scores(self, topic: str, payload) -> None:
        # Unpack message batch
        batch_msg = HydraMsg.from_dict(payload)
        messages = batch_msg.payload[DHydraMQ.MSGS]
        self._cur_scores_batch_num = messages[0]["payload"]["count"]
        if self._first_scores_batch:
            self._first_scores_batch_num = False
            self._prev_scores_batch_num = self._cur_scores_batch_num - 1
        if (self._cur_scores_batch_num - self._prev_scores_batch_num) != 1:
            self.event_log.add_event(
                ev_type=DField.WARNING, event="Score data is being dropped"
            )
        self._prev_scores_batch_num = self._cur_scores_batch_num

        metrics = self.metrics

        for payload in messages[1:]:

            # Get the data from the payload
            cur_epoch = payload.get(DGameField.EPOCH)
            cur_score = payload.get(DGameField.CUR_SCORE)
            highscore = payload.get(DGameField.HIGHSCORE)
            final_score = payload.get(DNetField.FINAL_SCORE)

            # Load the data into the metrics object
            metrics.add_cur_epoch(cur_epoch)
            metrics.add_cur_score(cur_score)
            is_new_highscore = metrics.add_highscore(highscore)
            if final_score is not None:
                is_new_final_score = metrics.add_final_score(final_score)
                if is_new_final_score:
                    self.telemtry.game_score_plot.plot_cur_scores()
                    self.telemtry.scores_dist_plot.plot_all()

            ## Update the TUI

            # Only update if the highscore is new
            if is_new_highscore:
                highscore_log = self.query_one(
                    f"#{DField.HIGHSCORES_LOG}", HighScoresLog
                )
                highscore_log.border_subtitle = (
                    f"{DLabel.HIGHSCORE}: {highscore}"
                )
                highscore_event = metrics.get_last_highscore_event()
                highscore_log.add_highscore(highscore_event)
                self.event_log.add_event(
                    ev_type=DField.HIGHSCORE,
                    event=f"New highscore: {highscore}",
                )
                self.telemtry.game_score_plot.plot_highscores()

            # Only update the score, if "per_step" is enabled.
            if self.cfg.get(DNetField.PER_STEP):
                self._w_board_box.border_subtitle = (
                    f"{DLabel.SCORE}: {cur_score:<2}"
                )

    async def on_shutdown_request(self) -> None:
        if self.mq is not None:
            await self.mq.quit()
            self.mq = None

    def on_select_changed(self, event: Select.Changed) -> None:
        if event.control.id != DField.MODEL_TYPE_SELECT:
            return

        model_type = event.value
        # Linear model defaults
        if model_type == DField.LINEAR:
            batch_size = DLinear.BATCH_SIZE
            dropout_p = DLinear.DROPOUT_P
            epsilon_decay = DLinear.EPSILON_DECAY_RATE
            gamma = DLinear.GAMMA
            hidden_size = DLinear.HIDDEN_SIZE
            initial_epsilon = DLinear.INITIAL_EPSILON
            lr = f"{DLinear.LEARNING_RATE:.5f}"
            min_epsilon = DLinear.MINIMUM_EPSILON
            self.remove_class(DField.RNN)
            self.add_class(DField.LINEAR)

        # RNN model defaults
        elif model_type == DField.RNN:
            batch_size = DRNN.BATCH_SIZE
            dropout_p = DRNN.DROPOUT_P_VALUE
            epsilon_decay = DRNN.EPSILON_DECAY_RATE
            gamma = DRNN.GAMMA
            hidden_size = DRNN.HIDDEN_SIZE
            initial_epsilon = DRNN.INITIAL_EPSILON
            lr = f"{DRNN.LEARNING_RATE:.4f}"
            min_epsilon = DRNN.MINIMUM_EPSILON
            self.remove_class(DField.LINEAR)
            self.add_class(DField.RNN)

        else:
            raise ValueError(f"EFFOR: Unrecognized model type {model_type}")

        # Update widgets with model specific defaults
        self._w_batch_size_input.value = str(batch_size)
        self._w_dropout_p_input.value = str(dropout_p)
        self._w_epsilon_decay_input.value = str(epsilon_decay)
        self._w_gamma_input.value = str(gamma)
        self._w_hidden_size_input.value = str(hidden_size)
        self._w_initial_epsilon_input.value = str(initial_epsilon)
        self._w_learning_rate_input.value = lr
        self._w_min_epsilon_input.value = str(min_epsilon)

        # Update HydraMetrics
        self.metrics.set_initial_epsilon(initial_epsilon)

    async def _on_sim_event(self, topic: str, payload) -> None:
        self.event_log.add_event(
            ev_type=payload[DEvent.SENDER], event=payload[DEvent.MESSAGE]
        )

    async def on_switch_changed(self, event: Switch.Changed) -> None:
        if event.control.id != DField.TURBO_MODE:
            return

        # Turbo enabled, replaces the Input with a Label showing "0.0"
        if event.value:
            self.remove_class(DField.TURBO_OFF)
            self.add_class(DField.TURBO_ON)
            self.event_log.add_event(
                ev_type=DField.SIM_LOOP,
                event="Turbo mode enabled: Game rendering disabled, move delay set to 0.0",
            )
            self.mq.disable_per_step_sub()
        # Turbo disabled
        else:
            self.remove_class(DField.TURBO_ON)
            self.add_class(DField.TURBO_OFF)
            move_delay = self.query_one(f"#{DField.MOVE_DELAY_INPUT}").value
            self.event_log.add_event(
                ev_type=DField.SIM_LOOP,
                event=f"Turbo mode disabled: Game rendering enabled, move delay set to {move_delay}",
            )
            self.mq.enable_per_step_sub()

        self._update_tui_labels()
        msg = HydraMsg(
            sender=DModule.HYDRA_CLIENT,
            target=DModule.HYDRA_MGR,
            method=DGameMethod.UPDATE_CONFIG,
            payload=self.cfg.to_dict(),
        )
        await self.mq.send(msg)

    async def _send_handshake(self):
        msg = HydraMsg(
            sender=DModule.HYDRA_CLIENT,
            target=DModule.HYDRA_MGR,
            method=DMethod.HANDSHAKE,
        )
        await self.mq.send(msg)

        try:
            reply = await self.mq.recv()
            sim_running = reply.payload.pop(DGameField.SIM_RUNNING)

            # ----- Simulation is running ---
            if sim_running:
                # Always receive per-episode and event updates
                self.mq.enable_per_episode_sub()
                self.mq.enable_events_sub()

                # Always receive score updates
                self.mq.enable_scores_sub()
                self.add_class(DField.SIM_RUNNING)
                self.cfg = SimCfg.from_dict(reply.payload)

                # Turbo mode disabled
                if self.cfg.get(DNetField.PER_STEP):
                    self.mq.enable_per_step_sub()
                    self.remove_class(DField.TURBO_ON)
                    self.add_class(DField.TURBO_OFF)
                    self.query_one(f"#{DField.TURBO_MODE}", Switch).value = (
                        False
                    )

                # Turbo mode enabled
                else:
                    self.mq.disable_per_step_sub()
                    self.remove_class(DField.TURBO_OFF)
                    self.add_class(DField.TURBO_ON)
                    self.query_one(f"#{DField.TURBO_MODE}", Switch).value = (
                        True
                    )

                # Load configurable TUI settings from the running sim config
                self._w_initial_epsilon_input.value = str(
                    self.cfg.get(DNetField.INITIAL_EPSILON)
                )
                self._w_min_epsilon_input.value = str(
                    self.cfg.get(DNetField.MIN_EPSILON)
                )
                self._w_move_delay_input.value = str(
                    self.cfg.get(DNetField.MOVE_DELAY)
                )
                self._w_model_type_select.value = self.cfg.get(
                    DNetField.MODEL_TYPE
                )
                self._w_hidden_size_input.value = str(
                    self.cfg.get(DNetField.HIDDEN_SIZE)
                )
                self._update_tui_labels()
                self.event_log.add_event(
                    ev_type=DField.SIM_LOOP,
                    event="Connecting to running simulation",
                )

            # ----- Simulation is not running ---
            else:
                self.add_class(DField.SIM_STOPPED)
                self.event_log.add_event(
                    ev_type=DField.SIM_LOOP,
                    event="Connected to simulation server",
                )
            self.remove_class(DField.BAD_HANDSHAKE)

        # ----- Cannot connect to Simulation Server (HydraMgr)
        except asyncio.TimeoutError:
            self.event_log.add_event(
                ev_type=DField.SIM_LOOP,
                event="Unable to connect to simulation server",
            )

    async def _send_reset(self):
        msg = HydraMsg(
            sender=DModule.HYDRA_CLIENT,
            target=DModule.HYDRA_MGR,
            method=DGameMethod.RESET_GAME,
        )
        await self.mq.send(msg)

        try:
            reply = await self.mq.recv()
        except asyncio.TimeoutError:
            pass

    async def _send_start_run(self):
        # Send a START_RUN message to the HydraMgr
        msg = HydraMsg(
            sender=DModule.HYDRA_CLIENT,
            target=DModule.HYDRA_MGR,
            method=DGameMethod.START_RUN,
            payload=self.cfg.to_dict(),
        )
        await self.mq.send(msg)

        # Check what the HydraMgr sent back:
        try:
            reply = await self.mq.recv()
            if reply.payload[DGameField.OK]:
                # Perfect, sim wasn't running now it is.
                self.remove_class(DField.SIM_STOPPED)
                self.add_class(DField.SIM_RUNNING)
            else:
                # Sim was already running
                self.remove_class(DField.SIM_STOPPED)
                self.add_class(DField.SIM_RUNNING)
            self.event_log.add_event(
                ev_type=DField.SIM_LOOP, event="Simulation started"
            )

        except asyncio.TimeoutError:
            pass

    async def _send_stop_run(self):
        msg = HydraMsg(
            sender=DModule.HYDRA_CLIENT,
            target=DModule.HYDRA_MGR,
            method=DGameMethod.STOP_RUN,
        )
        await self.mq.send(msg)
        try:
            reply = await self.mq.recv()

        except asyncio.TimeoutError:
            pass

    async def _send_update_config(self):
        msg = HydraMsg(
            sender=DModule.HYDRA_CLIENT,
            target=DModule.HYDRA_MGR,
            method=DGameMethod.UPDATE_CONFIG,
            payload=self.cfg.to_dict(),
        )
        await self.mq.send(msg)
        if self.query_one(f"#{DField.TURBO_MODE}", Switch).value:
            self.mq.disable_per_step_sub()
        else:
            self.mq.enable_per_step_sub()

    async def _take_snapshot(self):
        # Snapshot file
        now = datetime.now()
        timestamp = now.strftime("%Y-%m-%d_%H-%M.txt")
        snapshot_file = DFile.BASE_SNAPSHOT + timestamp
        fq_snapshot = os.path.join(self._hydra_dir, snapshot_file)
        self.snapshot.create_snapshot(snap_file=fq_snapshot, cfg=self.cfg)

        self.event_log.add_event(
            ev_type=DField.SNAPSHOT,
            event=f"Created snapshot file: {fq_snapshot}",
        )

    def _update_tui_labels(self):
        """
        Update the SimCfg settings and the TUI labels
        """
        # Batch size
        batch_size = self._w_batch_size_input.value
        self._w_batch_size_label.update(batch_size)
        # Random Seed
        random_seed = self._w_random_seed_input.value
        self._w_random_seed_label.update(random_seed)
        # Epsilon values
        epsilon_decay = self._w_epsilon_decay_input.value
        initial_epsilon = self._w_initial_epsilon_input.value
        min_epsilon = self._w_min_epsilon_input.value
        self._w_epsilon_decay_label.update(str(epsilon_decay))
        self._w_initial_epsilon_label.update(str(initial_epsilon))
        self._w_min_epsilon_label.update(str(min_epsilon))
        # Turbo mode == Not per_step
        per_step = not self.query_one(f"#{DField.TURBO_MODE}", Switch).value
        # Move delay
        move_delay = self._w_move_delay_input.value
        # Model type
        model_type = self._w_model_type_select.value
        model_type_label = MODEL_TYPE_TABLE[model_type]
        self._w_model_type_label.update(model_type_label)
        # Hidden size
        hidden_size = self._w_hidden_size_input.value
        self._w_hidden_size_label.update(str(hidden_size))
        # Learning rate
        learning_rate = self._w_learning_rate_input.value
        self._w_learning_rate_label.update(learning_rate)
        # Discount/Gamma
        gamma = self._w_gamma_input.value
        self._w_gamma_label.update(gamma)
        # Dropout p-value
        dropout_p = self._w_dropout_p_input.value
        self._w_dropout_p_label.update(dropout_p)
        # RNN layers
        rnn_layers = self._w_rnn_layers_input.value
        self._w_rnn_layers_label.update(rnn_layers)
        # RNN Tau
        rnn_tau = self._w_rnn_tau_input.value
        self._w_rnn_tau_label.update(rnn_tau)
        # Sequence length
        seq_length = self._w_seq_length_input.value
        self._w_seq_length_label.update(seq_length)

        self.cfg.apply(
            {
                DNetField.BATCH_SIZE: batch_size,
                DNetField.DROPOUT_P: dropout_p,
                DNetField.EPSILON_DECAY: epsilon_decay,
                DNetField.GAMMA: gamma,
                DNetField.HIDDEN_SIZE: hidden_size,
                DNetField.INITIAL_EPSILON: initial_epsilon,
                DNetField.LEARNING_RATE: learning_rate,
                DNetField.MIN_EPSILON: min_epsilon,
                DNetField.PER_STEP: per_step,
                DNetField.MODEL_TYPE: model_type,
                DNetField.MOVE_DELAY: move_delay,
                DNetField.RANDOM_SEED: random_seed,
                DNetField.RNN_LAYERS: rnn_layers,
                DNetField.RNN_TAU: rnn_tau,
                DNetField.SEQ_LENGTH: seq_length,
            }
        )


def main() -> None:
    tui = HydraClientTui()
    tui.run()


if __name__ == "__main__":
    main()



# ai_hydra/constants/DGame.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0


from typing import Final


class DGameDef:
    BOARD_HEIGHT: Final[int] = 20
    BOARD_WIDTH: Final[int] = 20
    INITIAL_SNAKE_LENGTH: Final[int] = 3
    # MOVE_BUDGET: Final[int] = 100
    MAX_MOVES_MULTIPLIER: Final[int] = 100
    FOOD_REWARD: Final[int] = 10
    COLLISION_PENALTY: Final[int] = -10
    EMPTY_MOVE_REWARD: Final[int] = 0.0
    CLOSER_TO_FOOD: Final[float] = 1.5
    FURTHER_FROM_FOOD: Final[float] = -1.5


class DGameField:
    ACTION: Final[str] = "action"
    BOARD: Final[str] = "board"
    CLOSER_TO_FOOD: Final[str] = "closer_to_food"
    CUR_SCORE: Final[str] = "cur_score"
    CUR_EPOCH: Final[str] = "cur_epoch"
    DIRECTION: Final[str] = "direction"
    DONE: Final[str] = "done"
    DX: Final[str] = "dx"
    DY: Final[str] = "dy"
    EMPTY: Final[str] = "empty"
    ERROR: Final[str] = "error"
    ELAPSED_TIME: Final[str] = "elapsed_time"
    EPISODE_DONE: Final[str] = "episode_done"
    EPISODE_ID: Final[str] = "episode_id"
    EPOCH: Final[str] = "epoch"
    FOOD: Final[str] = "food"
    FOOD_POSITION: Final[str] = "food_position"
    FURTHER_FROM_FOOD: Final[str] = "further_from_food"
    GRID_SIZE: Final[str] = "grid_size"
    H: Final[str] = "h"
    HIGHSCORE: Final[str] = "highscore"
    HIGHSCORE_NLH: Final[str] = "highscore_nlh"
    HIGHSCORE_LH: Final[str] = "highscore_lh"
    HIGHSCORE_EVENT: Final[str] = "highscore_event"
    HIGHSCORE_EVENT_LH: Final[str] = "highscore_event_lh"
    HIGHSCORE_EVENT_NLH: Final[str] = "highscore_event_nlh"
    INFO: Final[str] = "info"
    INVALID_ACTION: Final[str] = "invalid_action"
    LEFT_TURN: Final[str] = "left_turn"
    MAX_MOVES: Final[str] = "max_moves"
    MISSING_ACTION: Final[str] = "missing_action"
    MOVE_COUNT: Final[str] = "move_count"
    OK: Final[str] = "ok"
    PUB_TYPE: Final[str] = "pub_type"
    REASON: Final[str] = "reason"
    REWARD: Final[str] = "reward"
    REWARD_TOTAL: Final[str] = "reward_total"
    RIGHT_TURN: Final[str] = "right_turn"
    SCORE: Final[str] = "score"
    SCORE_DELTA: Final[str] = "score_delta"
    SEED: Final[str] = "seed"
    SESSION_ID: Final[str] = "session_id"
    SIM_RUNNING: Final[str] = "sim_running"
    SNAKE: Final[str] = "snake"
    SNAKE_BODY: Final[str] = "snake_body"
    SNAKE_HEAD: Final[str] = "snake_head"
    SNAPSHOT: Final[str] = "snapshot"
    STEP_N: Final[str] = "step_n"
    STRAIGHT: Final[str] = "straight"
    WALL: Final[str] = "wall"
    W: Final[str] = "w"
    X: Final[str] = "x"
    Y: Final[str] = "y"


class DGameLabel:
    NO_FREE_POSITIONS: Final[str] = "No free positions available for food"


class DGameMethod:
    RESET_GAME: Final[str] = "reset_game"
    START_RUN: Final[str] = "start_run"
    XXX_GAME_STEP: Final[str] = "game_step"
    XXX_PUB_TYPE: Final[str] = "pub_type"
    STOP_RUN: Final[str] = "stop_run"
    XXX_UPDATE: Final[str] = "update"
    UPDATE_CONFIG: Final[str] = "update_config"



# ai_hydra/utils/HydraEventMQ.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0
#

from typing import Callable, Awaitable, Any
from dataclasses import dataclass, field

import json
import zmq
import zmq.asyncio

from ai_hydra.constants.DHydra import DHydraLog
from ai_hydra.constants.DHydraMQ import DEvent, DHydraMQ, DHydraMQDef


@dataclass(slots=True, frozen=True)
class EventMsg:
    level: DHydraLog
    message: str
    payload: dict[Any, Any] = field(default_factory=dict)


class HydraEventMQ:

    def __init__(self, client_id: str, pub_func):
        self._client_id = client_id
        self._pub_event = pub_func

    async def publish(self, event: EventMsg):
        msg_dict = {
            DEvent.SENDER: self._client_id,
            DEvent.LEVEL: event.level,
            DEvent.MESSAGE: event.message,
            DEvent.PAYLOAD: event.payload,
        }
        await self._pub_event(msg_dict)



# ai_hydra/constants/DReplayMemory.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from typing import Final


class DMemory:
    MAX_CHUNKS: Final[int] = 3000  # Used by RNN
    MIN_CHUNKS: Final[int] = 100  # Used by RNN
    MIN_FRAMES: Final[int] = 1500  # Used by Linear
    MAX_MEM_SIZE: Final[int] = 50000  # Used by Linear



# ai_hydra/constants/DEvent.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from ai_hydra.constants.DHydraTui import DField
from ai_hydra.constants.DHydra import DModule


EVENT_MAP = {
    DField.SIM_LOOP: "⚡ Simulation",
    DField.SNAPSHOT: "📸 Snapshot",
    DField.TUI: "💻 Hydra Client",
    DField.WARNING: "⚠️ Warning",
    DModule.EPSILON_ALGO: "🎲 Epsilon",
}



# ai_hydra/utils/HydraHelper.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0


# Helper function
def minutes_to_uptime(seconds: int):
    # Return a string like:
    # 45s
    # 7h 23m
    # 1d 7h 32m
    days, minutes = divmod(int(seconds), 86400)
    hours, minutes = divmod(minutes, 3600)
    minutes, seconds = divmod(minutes, 60)

    if days > 0:
        if seconds < 10:
            seconds = f" {seconds}"
        if minutes < 10:
            minutes = f" {minutes}"
        if hours < 10:
            hours = f" {hours}"
        return f"{days}d {hours}h {minutes}m"

    elif hours > 0:
        if seconds < 10:
            seconds = f" {seconds}"
        if minutes < 10:
            minutes = f" {minutes}"
        return f"{hours}h {minutes}m"

    elif minutes > 0:
        if seconds < 10:
            seconds = f" {seconds}"
        if minutes < 10:
            return f" {minutes}m {seconds}s"
        return f"{minutes}m {seconds}s"

    else:
        return f"{seconds}s"



# ai_hydra/utils/DBMgr.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

import os
import random
import sqlite3, pickle
from pathlib import Path

from ai_hydra.constants.DHydra import DHydra, DHydraLog
from ai_hydra.constants.DReplayMemory import DMemory
from ai_hydra.constants.DHydraTui import DFile

from ai_hydra.utils.HydraLog import HydraLog


class DBMgr:

    def __init__(self, log_level: DHydraLog):

        self.log = HydraLog(
            client_id="DBMgr", log_level=log_level, to_console=True
        )

        # We need determinism for RL experiments
        random.seed(DHydra.RANDOM_SEED)

        # The minimum number of episodes, before anyting is returned from
        # the replay memory.
        self._min_games = DMemory.MIN_GAMES
        # How large batches of frames should be
        self._batch_size = DMemory.BATCH_SIZE  # 256

        # AI Hydra stores temporary and persistent files in this directory
        ai_hydra_dir = os.path.join(Path.home(), DHydra.HYDRA_DIR)
        if not os.path.exists(ai_hydra_dir):
            os.mkdir(ai_hydra_dir)
        self._db_file = os.path.join(ai_hydra_dir, DFile.HYDRA_SERVER_DB)

        # Start with a new DB file (no persistence)
        if os.path.exists(self._db_file):
            os.remove(self._db_file)

        # Connect to SQLite
        self._conn = sqlite3.connect(self._db_file, check_same_thread=False)

        # Get a cursor
        self._cursor = self._conn.cursor()

        # Initialize the DB
        self._init_db()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Try to ensure the DB shuts down cleanly.
        """
        self.close()

    def __del__(self):
        """
        Try to ensure the DB shuts down cleanly.
        """
        try:
            self.close()
        except Exception:
            pass

    def add_game(self, final_score, total_frames):
        """
        Add a game into the games table.
        """
        self._cursor.execute(
            "INSERT INTO games (score, total_frames) VALUES (?, ?)",
            (final_score, total_frames),
        )
        self._conn.commit()
        return self._cursor.lastrowid  # game_id

    def clear_runtime_data(self) -> None:
        self._cursor.executescript(
            """
            DELETE from games;

            DELETE from frames;
            """
        )
        self._conn.commit()

    def close(self):
        """
        Close the database.
        """
        if getattr(self, "_conn", None):
            self._conn.close()
            self._conn = None
        if getattr(self, "_db_file", None) and os.path.exists(self._db_file):
            os.remove(self._db_file)
            self._db_file = None

    def cursor(self) -> None:
        """
        Return a cursor (used by ReplayMemory).
        """
        return self._cursor

    def get_avg_game_length(self):
        self._cursor.execute("SELECT AVG(total_frames) FROM games")
        avg = self._cursor.fetchone()[0]
        return int(avg) if avg else 0

    def get_random_frames(self, batch_size=None):
        """Return a random set of frames from the database. Do not use the SQLite3
        RANDOM() function, because we can't set the seed."""

        if batch_size:
            num_frames = batch_size
        else:
            num_frames = (
                self.get_avg_game_length() or 32
            )  # fallback if no data

        # Retrieve the id values from the frames table
        self._cursor.execute("SELECT id FROM frames")
        all_ids = [row[0] for row in self._cursor.fetchall()]

        # if len(all_ids) < self._batch_size * 10:
        if len(all_ids) < 1000:
            return [], 0

        # Get n random ids
        sample_ids = random.sample(all_ids, min(num_frames, len(all_ids)))
        if not sample_ids:
            return [], 0

        # Build placeholders for the SQL IN clause
        placeholders = ",".join("?" for _ in sample_ids)

        # Execute the query with the unpacked tuple
        self._cursor.execute(
            f"SELECT state, action, reward, next_state, done FROM frames "
            f"WHERE id IN ({placeholders}) ORDER BY id ASC",
            sample_ids,
        )
        rows = self._cursor.fetchall()
        frames = [
            (
                pickle.loads(state_blob),
                pickle.loads(action),
                float(reward),
                pickle.loads(next_state_blob),
                bool(done),
            )
            for state_blob, action, reward, next_state_blob, done in rows
        ]
        return frames, len(frames)

    def get_random_game(self):
        self._cursor.execute("SELECT id FROM games")
        all_ids = [row[0] for row in self._cursor.fetchall()]
        if not all_ids or len(all_ids) < DMemory.MIN_GAMES:
            return [], DMemory.NO_DATA  # no games available

        rand_id = random.choice(all_ids)
        self._cursor.execute(
            "SELECT state, action, reward, next_state, done "
            "FROM frames WHERE game_id = ? ORDER BY frame_index ASC",
            (rand_id,),
        )
        rows = self._cursor.fetchall()
        if not rows:
            return [], rand_id  # game exists but no frames

        game = [
            (
                pickle.loads(state_blob),
                pickle.loads(action),
                float(reward),
                pickle.loads(next_state_blob),
                bool(done),
            )
            for state_blob, action, reward, next_state_blob, done in rows
        ]
        return game, rand_id

    def _init_db(self) -> None:
        # Create the tables
        self._cursor.executescript(
            """
            CREATE TABLE IF NOT EXISTS games (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                score INTEGER NOT NULL,
                total_frames INTEGER NOT NULL
            );

            CREATE TABLE IF NOT EXISTS frames (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                game_id INTEGER NOT NULL,
                frame_index INTEGER NOT NULL,
                state BLOB NOT NULL,
                action BLOB NOT NULL,      
                reward INTEGER NOT NULL,
                next_state BLOB NOT NULL,
                done INTEGER NOT NULL,        -- 0 or 1
                FOREIGN KEY (game_id) REFERENCES games(id)
            );

            CREATE UNIQUE INDEX IF NOT EXISTS idx_game_frame ON frames (game_id, frame_index);

            CREATE INDEX IF NOT EXISTS idx_frames_game_id ON frames (game_id);
            """
        )
        self._conn.commit()

        # If the games or frames tables do exit, clear the data
        self.clear_runtime_data()



# ai_hydra/nnet/ReplayMemory.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0


from __future__ import annotations

from random import Random
from dataclasses import dataclass, field

from ai_hydra.constants.DReplayMemory import DMemory
from ai_hydra.constants.DHydra import DHydraLog

from ai_hydra.nnet.Transition import Transition
from ai_hydra.utils.HydraLog import HydraLog

MAX_FRAMES = DMemory.MAX_MEM_SIZE  # Max Linear transitions
MAX_BUCKETS = 20
THRESHOLDS_REQUIRED = 5

# Gear => (promotion_score, seq_length, batch_size)
GEARBOX = {
    1: (10, 4, 128),
    2: (25, 8, 64),
    3: (40, 16, 32),
    4: (55, 32, 16),
    5: (70, 64, 8),
}


@dataclass(slots=True, frozen=True)
class GearMeta:
    """
    This stores, for one episode and under one gear:

    - Ssequence length
    - Number of chunks
    - Slice ranges for each bucket/chunk
    """

    seq_length: int
    num_chunks: int
    ranges: tuple[tuple[int, int], ...]


@dataclass(slots=True, frozen=True)
class Episode:
    """
    This stores the frames for one game and additional metadata.
    """

    frames: tuple[Transition, ...]
    size: int
    gear_meta: dict[int, GearMeta] = field(default_factory=dict)


class ATHReplayMemory:
    def __init__(
        self,
        rng: Random,
        log_level: DHydraLog,
    ):
        self.log = HydraLog(
            client_id="ReplayMemory",
            log_level=log_level,
            to_console=True,
        )
        self._rng = rng

        self._episodes_by_bucket: dict[int, list[Episode]] = {
            idx: [] for idx in range(MAX_BUCKETS)
        }
        self._warmed_buckets: list[int] = []

        self._cur_game: list[Transition] = []
        self._games: list[Episode] = []
        self._stored_frames = 0

        self._cur_gear = 1
        _, self._cur_seq_length, self._batch_size = GEARBOX[self._cur_gear]

        self._threshold_achieved_count = 0

        self._has_logged_pruning = False

        self.log.info("Initialized")
        self.log.info(f"Setting max_frames to {MAX_FRAMES}")
        self.log.info(f"Setting max_buckets to {MAX_BUCKETS}")

    def append(self, t: Transition, final_score: int | None = None) -> None:
        self._cur_game.append(t)

        if t.done:
            if final_score is None:
                raise ValueError(
                    "final_score must be provided when appending a terminal transition"
                )

            self._finalize_game()
            self._prune_if_needed()

            shifted = self._check_gear(final_score)
            self._rebuild_bucket_index()

    def _build_episode_gear_meta(
        self,
        ep_size: int,
    ) -> dict[int, GearMeta]:
        """
        Build metadata for all gears for one episode.
        """
        gear_meta: dict[int, GearMeta] = {}

        for gear, (_, seq_length, _) in GEARBOX.items():
            gear_meta[gear] = self._build_gear_meta_for_size(
                ep_size=ep_size,
                seq_length=seq_length,
            )

        return gear_meta

    def _build_gear_meta_for_size(
        self,
        ep_size: int,
        seq_length: int,
    ) -> GearMeta:
        """
        Build gear-specific chunk metadata for an episode size.

        Chunks are aligned from the end, matching the old ReplayMemory behavior.
        """
        num_chunks = ep_size // seq_length
        rem = ep_size % seq_length

        ranges: list[tuple[int, int]] = []

        for chunk_idx in range(num_chunks):
            start = rem + (chunk_idx * seq_length)
            end = start + seq_length
            ranges.append((start, end))

        return GearMeta(
            seq_length=seq_length,
            num_chunks=num_chunks,
            ranges=tuple(ranges),
        )

    def _check_gear(self, final_score: int) -> bool:
        if self._cur_gear == max(GEARBOX):
            return False

        promotion_score, _, _ = GEARBOX[self._cur_gear]

        if final_score < promotion_score:
            return False

        self._threshold_achieved_count += 1

        if self._threshold_achieved_count < THRESHOLDS_REQUIRED:
            return False

        self._cur_gear += 1
        self._threshold_achieved_count = 0
        _, self._cur_seq_length, self._batch_size = GEARBOX[self._cur_gear]

        self.log.info(
            f"Shifting into higher gear: {self._cur_gear} "
            f"(seq_length={self._cur_seq_length}, batch_size={self._batch_size})"
        )

        return True

    def _chunk_range_for_bucket(
        self, ep_size: int, seq_length: int, bucket_idx: int
    ) -> tuple[int, int] | None:
        pass

    def get_batch_size(self) -> int:
        return self._batch_size

    def get_seq_length(self) -> int:
        return self._cur_seq_length

    def get_current_gear(self) -> int:
        return self._cur_gear

    def get_warmed_buckets(self) -> list[int]:
        return list(self._warmed_buckets)

    def _finalize_game(self) -> None:
        frames = tuple(self._cur_game)
        self._cur_game = []

        ep_size = len(frames)
        self._stored_frames += ep_size

        gear_meta = self._build_episode_gear_meta(ep_size)

        self._games.append(
            Episode(
                frames=frames,
                size=ep_size,
                gear_meta=gear_meta,
            )
        )

    def _num_chunks_for_episode(self, ep_size: int, seq_length: int) -> int:
        return ep_size // seq_length

    def _prune_if_needed(self) -> None:
        while self._stored_frames > MAX_FRAMES:
            episode = self._games.pop(0)
            self._stored_frames -= episode.size

            if not self._has_logged_pruning:
                self.log.info("Memory is full, pruning initiated")
                self._has_logged_pruning = True

    def _rebuild_bucket_index(self) -> None:
        """
        Rebuild current-gear bucket eligibility index.
        """
        self._episodes_by_bucket = {idx: [] for idx in range(MAX_BUCKETS)}

        for ep in self._games:
            meta = ep.gear_meta[self._cur_gear]
            usable_buckets = min(meta.num_chunks, MAX_BUCKETS)

            for bucket_idx in range(usable_buckets):
                self._episodes_by_bucket[bucket_idx].append(ep)

        self._warmed_buckets = [
            idx
            for idx, episodes in self._episodes_by_bucket.items()
            if episodes
        ]

    def sample_chunks(self) -> list[list[Transition]] | None:
        """
        Sample chunks across warmed buckets for the current gear.
        """
        if not self._warmed_buckets:
            return None

        total_chunks = 0
        for bucket_idx in self._warmed_buckets:
            total_chunks += len(self._episodes_by_bucket[bucket_idx])

        if total_chunks < self._batch_size:
            return None

        samples: list[list[Transition]] = []

        base = self._batch_size // len(self._warmed_buckets)
        remainder = self._batch_size % len(self._warmed_buckets)

        requested: dict[int, int] = {idx: base for idx in self._warmed_buckets}

        warmed_desc = sorted(self._warmed_buckets, reverse=True)
        for i in range(remainder):
            requested[warmed_desc[i % len(warmed_desc)]] += 1

        for bucket_idx in self._warmed_buckets:
            eligible = self._episodes_by_bucket[bucket_idx]
            if not eligible:
                continue

            meta_samples = requested[bucket_idx]

            for _ in range(meta_samples):
                ep = self._rng.choice(eligible)
                meta = ep.gear_meta[self._cur_gear]
                start, end = meta.ranges[bucket_idx]
                samples.append(list(ep.frames[start:end]))

        return samples



# ai_hydra/server/SnakeMgr.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from __future__ import annotations

import random
from dataclasses import dataclass
from typing import Any, Optional
from datetime import datetime
import traceback

from ai_hydra.constants.DHydra import DHydra
from ai_hydra.constants.DGame import DGameField
from ai_hydra.constants.DNNet import DNetField

from ai_hydra.game.GameLogic import GameLogic
from ai_hydra.nnet.Policy.HydraPolicy import HydraPolicy
from ai_hydra.utils.SimCfg import SimCfg
from ai_hydra.utils.HydraLog import HydraLog

from ai_hydra.utils.HydraHelper import minutes_to_uptime


@dataclass
class SnakeSession:
    """
    Pure domain session state for one client identity.
    """

    board: Any  # GameBoard, but keep Any here to avoid import cycles if needed
    seed: int
    rng: random.Random
    # Attribs with defaults
    done: bool = False
    episode_id: int = 0
    epoch: int = 0
    highscore: int = 0  # Actual highscore
    reward_total: float = 0.0
    score: int = 0
    step_n: int = 0


class SnakeMgr:
    """
    Domain manager for Snake game sessions.

    Owns:
      - a master seed + seed RNG (for minting per-session seeds)
      - a per-client session map

    Does NOT:
      - talk to MQ
      - render
      - run neural nets (future)
    """

    def __init__(
        self,
        cfg: SimCfg,
        log_level: HydraLog,
        master_seed: int = DHydra.RANDOM_SEED,
    ) -> None:
        if cfg is None:
            raise TypeError("SnakeMgr requires cfg (SimCfg)")
        self.cfg = cfg
        self.log = HydraLog(
            client_id="SnakeMgr",
            log_level=log_level,
            to_console=True,
        )

        self.master_seed = int(master_seed)

        self.seed_rng = random.Random(self.master_seed)
        self.sessions: dict[str, SnakeSession] = {}
        self.policy: Optional[HydraPolicy] = None
        self._start_time = datetime.now()

    # -------------------------
    # RNG API (Pattern B)
    # -------------------------

    def new_rng(self, seed: Optional[int] = None) -> tuple[int, random.Random]:
        """
        Create a brand-new RNG stream for a session.

        If seed is None, mint a deterministic seed from seed_rng.
        Returns (seed, rng).
        """
        if seed is None:
            seed = self.seed_rng.getrandbits(64)
        else:
            seed = int(seed)

        return seed, random.Random(seed)

    def clone_rng(self, parent_rng: random.Random) -> random.Random:
        """
        Clone RNG state exactly (for future clone rollouts).
        """
        clone = random.Random()
        clone.setstate(parent_rng.getstate())
        return clone

    def get_rng_state(self, rng: random.Random) -> object:
        """
        Export RNG state (useful later for replay/persistence).
        """
        return rng.getstate()

    def set_rng_state(self, rng: random.Random, state: object) -> None:
        """
        Restore RNG state (useful later for replay/persistence).
        """
        rng.setstate(state)

    # -------------------------
    # Session lifecycle
    # -------------------------

    def has_session(self, client_id: str) -> bool:
        return client_id in self.sessions

    # -------------------------
    # Game operations (Phase 1)
    # -------------------------

    def reset_session(
        self, client_id: str, seed: Optional[int] = None
    ) -> SnakeSession:
        from ai_hydra.game.GameBoard import GameBoard

        prev = self.sessions.get(client_id)
        prev_highscore = getattr(prev, "highscore", 0)
        prev_epoch = getattr(prev, "epoch", 0)
        next_step_n = getattr(prev, "step_n", 0) + 1

        session_seed, rng = self.new_rng(seed)
        episode_id = rng.getrandbits(32)

        board = GameBoard.new_game(
            rng=rng,
            seed=session_seed,
            episode_id=episode_id,
        )

        sess = SnakeSession(
            seed=session_seed,
            rng=rng,
            board=board,
            done=False,
            score=0,
            episode_id=episode_id,
            # carry-over stats
            highscore=prev_highscore,
            epoch=prev_epoch,
            step_n=next_step_n,
        )
        self.sessions[client_id] = sess
        return sess

    def get_session(self, client_id: str) -> SnakeSession:
        """
        Fetch a session. Auto-creates one if missing (seed minted).
        """
        sess = self.sessions.get(client_id)
        if sess is None:
            sess = self.reset_session(client_id, seed=None)
        return sess

    def reset(self, client_id: str, seed: Optional[int] = None) -> None:
        """
        Reset and return snapshot.
        """
        self.reset_session(client_id, seed=seed)

    def start_time(self, start_time=None):
        if start_time is not None:
            self._start_time = start_time
        return self._start_time

    def step(self, client_id: str, action: int) -> tuple[
        dict[str, Any],
        dict[str, Any],
        dict[str, Any] | None,
        dict[str, Any] | None,
    ]:
        """
        Step the client's session forward by one action.

        Returns:
        - state_dict
        """
        try:
            # The previous session....
            sess = self.get_session(client_id)

            # NN-ready state vector (pre-step)
            state = sess.board.get_state()

            # Apply action
            result = GameLogic.step(sess.board, int(action), sess.rng)

            sess.board = result.new_board

            reward = int(result.reward)
            done = bool(result.is_terminal)
            outcome = result.outcome  # use as "reason" for now

            # Score delta (Phase 1): +1 on FOOD, else 0
            if outcome == DGameField.FOOD:
                sess.score += 1

            elapsed_secs = (datetime.now() - self.start_time()).total_seconds()
            elapsed_str = minutes_to_uptime(int(elapsed_secs))

            sess.done = done
            sess.reward_total += reward

            # NN-ready next state (post-step)
            next_state = sess.board.get_state()

            # ----- State information ---
            state_dict = {
                DGameField.DONE: done,
                DNetField.STATE: state,
                DNetField.NEXT_STATE: next_state,
                DGameField.REWARD: reward,
                DGameField.REWARD_TOTAL: sess.reward_total,
                DGameField.EPISODE_ID: sess.episode_id,
            }

            # ----- The payload for the ZeroMQ "scores" topic ---
            scores_payload = {
                DGameField.EPOCH: sess.epoch,
                DGameField.CUR_SCORE: sess.score,
            }
            # Create a "final score" field for the TUI
            if done:
                scores_payload[DNetField.FINAL_SCORE] = sess.score

            # Add a "highscore" event
            if sess.score > sess.highscore:
                sess.highscore = sess.score
                scores_payload[DGameField.HIGHSCORE] = sess.highscore
                self.log.info(
                    f"Epoch: {sess.epoch} - New High Score: {sess.highscore}"
                )

            # ----- The payload for the ZeroMQ "per step" topic ---
            step_payload = None
            if self.cfg.get(DNetField.PER_STEP):
                step_payload = {DGameField.BOARD: sess.board.to_dict()}

            # ----- The payload for the ZeroMQ "per episode" topic ----
            ep_payload = None
            if sess.done:
                ep_payload: dict[str, Any] = {
                    DGameField.EPOCH: sess.epoch,
                    DGameField.REASON: outcome,
                    DGameField.ELAPSED_TIME: elapsed_str,
                }

            return state_dict, scores_payload, step_payload, ep_payload

        except Exception as e:
            self.log.critical(f"ERROR: {e}")
            self.log.critical(f"STACKTRACE: {traceback.format_exc()}")



# ai_hydra/constants/DNNet.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from typing import Final

from ai_hydra.constants.DGame import DGameField
from ai_hydra.constants.DHydraTui import DLabel, DField
import torch.nn as nn
import torch.optim as optim


class DRNN2:
    """
    RNN Model defaults
    """

    BATCH_SIZE: Final[int] = 24
    DROPOUT_P_VALUE: Final[float] = 0.05
    EPSILON_DECAY_RATE: Final[float] = 0.993
    GAMMA: Final[float] = 0.96  # Discount
    HIDDEN_SIZE: Final[int] = 448
    INITIAL_EPSILON: Final[float] = 0.999
    LEARNING_RATE: Final[float] = 0.0008
    MINIMUM_EPSILON: Final[float] = 0.005
    OUTPUT_SIZE: Final[int] = 3
    RNN_LAYERS: Final[int] = 5
    SEQ_LENGTH: Final[int] = 40
    TAU: Final[float] = 0.01


class DRNN:
    """
    RNN Model defaults
    """

    BATCH_SIZE: Final[int] = 24
    DROPOUT_P_VALUE: Final[float] = 0.05
    EPSILON_DECAY_RATE: Final[float] = 0.993
    GAMMA: Final[float] = 0.96  # Discount
    HIDDEN_SIZE: Final[int] = 224
    INITIAL_EPSILON: Final[float] = 0.999
    LEARNING_RATE: Final[float] = 0.0008
    MINIMUM_EPSILON: Final[float] = 0.005
    OUTPUT_SIZE: Final[int] = 3
    RNN_LAYERS: Final[int] = 3
    SEQ_LENGTH: Final[int] = 40
    TAU: Final[float] = 0.01


class DEpsilonField:
    """
    Epsilon-greedy fields.
    """

    INITIAL: Final[str] = "initial"
    MINIMUM: Final[str] = "minimum"
    DECAY_RATE: Final[str] = "decay"


class DLinear:
    """
    Constants for the Linear neural network
    """

    BATCH_SIZE: Final[int] = 64
    GAMMA: Final[float] = 0.9
    HIDDEN_SIZE: Final[int] = 192
    DROPOUT_P: Final[float] = 0.1
    LEARNING_RATE: Final[float] = 0.00005
    OUTPUT_SIZE: Final[int] = 3  # left / straight / right
    INITIAL_EPSILON: Final[float] = 0.99
    MINIMUM_EPSILON: Final[float] = 0.0
    EPSILON_DECAY_RATE: Final[float] = 0.98


class DNetDef:
    """
    Neural network constants.
    """

    # State map and hyperparameters
    DANGER_FEATURES: Final[int] = 3
    DIRECTION_FEATURES: Final[int] = 4
    FOOD_FEATURES: Final[int] = 2
    MOVE_DELAY: Final[float] = 0.02
    PER_STEP: Final[bool] = True
    STATE_LENGTH_BITS: Final[int] = 7

    # Derived
    INPUT_SIZE: Final[int] = (
        DANGER_FEATURES
        + DIRECTION_FEATURES
        + FOOD_FEATURES
        + STATE_LENGTH_BITS
    )


class DNetField:
    """
    Field names used in the NN
    """

    ACTION: Final[str] = DGameField.ACTION
    BATCH_SIZE: Final[str] = "batch_size"
    CUR_EPSILON: Final[str] = "cur_epsilon"
    DONE: Final[str] = DGameField.DONE
    DROPOUT_P: Final[str] = "dropout_p"
    EPSILON_DECAY: Final[str] = "epsilon_decay"
    FINAL_SCORE: Final[str] = "final_score"
    GAMMA: Final[str] = "gamma"
    HIDDEN_SIZE: Final[str] = "hidden_size"
    INITIAL_EPSILON: Final[str] = "initial_epsilon"
    LEARNING_RATE: Final[str] = "learning_rate"
    LOSS: Final[str] = "loss"
    MIN_EPSILON: Final[str] = "min_epsilon"
    MODEL_TYPE: Final[str] = "model_type"
    MOVE_DELAY: Final[str] = "move_delay"
    NEXT_STATE: Final[str] = "next_state"
    PER_STEP: Final[str] = "per_step"
    RANDOM_SEED: Final[str] = "random_seed"
    REWARD: Final[str] = DGameField.REWARD
    RNN_LAYERS: Final[str] = "rnn_layers"
    RNN_TAU: Final[str] = "rnn_tau"
    SEQ_LENGTH: Final[str] = "sequence_length"
    STATE: Final[str] = "state"


class DRNNTrainer:
    """
    RNN Trainer defaults
    """

    CRITERION = nn.SmoothL1Loss
    OPTIM = optim.Adam
    TAU: Final[float] = 0.005
    UPDATE_FREQ: Final[int] = 100


MODEL_TYPE_TABLE: Final[dict] = {
    DField.LINEAR: DLabel.LINEAR,
    DField.RNN: DLabel.RNN,
}

MODEL_TYPES: Final[list] = [
    (DLabel.LINEAR, DField.LINEAR),
    (DLabel.RNN, DField.RNN),
]



# ai_hydra/utils/HydraMetrics.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from collections import deque
from statistics import mean, median

from ai_hydra.constants.DHydraTui import DPlotDef
from ai_hydra.utils.MetricEvent import HighscoreEvent, ScoreEvent, LossEvent

MAX_CUR_SCORES = DPlotDef.MAX_CUR_SCORES
RECENT_SCORES_MAX = DPlotDef.RECENT_SCORES_MAX
RECENT_LOSS_MAX = DPlotDef.RECENT_LOSS_MAX

AVG_DIVISOR = 5
AVG_CUR_SCORES = MAX_CUR_SCORES // AVG_DIVISOR


class HydraMetrics:

    def __init__(self, initial_epsilon: float) -> None:
        self._cur_epsilon = initial_epsilon
        self._cur_epoch: int = 0
        self._cur_loss: float | None = None
        self._cur_score: int = 0
        self._elapsed_time: str = "0s"

        # Current scores data
        self._cur_scores: deque[ScoreEvent] = deque(maxlen=MAX_CUR_SCORES)
        self._avg_cur_scores: deque[ScoreEvent] = deque(maxlen=AVG_CUR_SCORES)
        self._avg_cur_scores_buf: list = []

        # Highscore events
        self._highscore_events: list[HighscoreEvent] = []

        # Score distribution
        self._scores_dist: dict[int, int] = {}
        self._recent_scores_dist: dict[int, int] = {}
        self._scores_dist_list: list[int] = []
        self._recent_scores_dist_list: list[int] = []

        # Loss events
        self._losses: list[LossEvent] = []
        self._recent_losses: deque[LossEvent] = deque(maxlen=RECENT_LOSS_MAX)

    def add_cur_epoch(self, epoch: int) -> None:
        self._cur_epoch = epoch

    def add_cur_epsilon(self, epsilon: float) -> None:
        self._cur_epsilon = epsilon

    def add_cur_loss(self, loss: float) -> None:
        if loss is None:
            return False

        # Don't insert dupes
        if self._losses and self._cur_epoch == self._losses[-1].epoch:
            return False

        self._cur_loss = loss
        self._losses.append(LossEvent(epoch=self._cur_epoch, loss=loss))
        self._recent_losses.append(LossEvent(epoch=self._cur_epoch, loss=loss))
        return True

    def add_cur_score(self, score: int) -> None:
        self._cur_score = score

    def add_elapsed_time(self, elapsed_time: str) -> None:
        self._elapsed_time = elapsed_time

    def add_final_score(self, score: int) -> None:
        cur_scores = self._cur_scores

        # Don't insert dupes
        if cur_scores and self._cur_epoch == cur_scores[-1].epoch:
            return False

        # Current Scores plot data
        score_event = ScoreEvent(epoch=self._cur_epoch, score=score)
        cur_scores.append(score_event)

        avg_buf = self._avg_cur_scores_buf
        avg_buf.append(score_event)
        if len(avg_buf) == AVG_DIVISOR:
            avg_epoch = avg_buf[-1].epoch
            avg_score = mean(e.score for e in avg_buf)
            self._avg_cur_scores.append(
                ScoreEvent(epoch=avg_epoch, score=avg_score)
            )
            self._avg_cur_scores_buf = []

        # All time score distribution data
        self._scores_dist[score] = self._scores_dist.get(score, 0) + 1
        self._scores_dist_list.append(score)

        # Recent time score distribution data
        self._recent_scores_dist[score] = (
            self._recent_scores_dist.get(score, 0) + 1
        )
        self._recent_scores_dist_list.append(score)

        if len(self._recent_scores_dist_list) > RECENT_SCORES_MAX:
            old_score = self._recent_scores_dist_list.pop(0)
            self._recent_scores_dist[old_score] -= 1
            if self._recent_scores_dist[old_score] == 0:
                del self._recent_scores_dist[old_score]

        return True

    def add_highscore(self, highscore: int) -> bool:
        """
        Add a highscore.

        Do not insert duplicate highscores. If the highscore parameter is a
        duplicate, return "False". Otherwise return "True".
        """
        if highscore is None:
            return False

        events = self._highscore_events
        # Don't insert dupes
        if events and highscore == events[-1].highscore:
            return False

        # Add new record
        events.append(
            HighscoreEvent(
                epoch=self._cur_epoch,
                highscore=highscore,
                epsilon=self._cur_epsilon,
                elapsed_time=self._elapsed_time,
            )
        )
        return True

    def get_avg_cur_score_plot_points(self) -> list[tuple[int, float]]:
        return [(e.epoch, e.score) for e in self._avg_cur_scores]

    def get_cur_epoch(self) -> int:
        return self._cur_epoch

    def get_cur_score(self) -> int:
        return self._cur_score

    def get_cur_score_plot_points(self) -> list[tuple[int, int]]:
        return [(e.epoch, e.score) for e in self._cur_scores]

    def get_elapsed_time(self) -> str:
        return self._elapsed_time

    def get_highscore_plot_points(self) -> list[tuple[int, int]]:
        return [(e.epoch, e.highscore) for e in self._highscore_events]

    def get_highscore_snapshot_rows(self) -> list[tuple[int, int, float, str]]:
        return [
            (e.epoch, e.highscore, e.epsilon, e.elapsed_time)
            for e in self._highscore_events
        ]

    def get_last_highscore_event(self) -> HighscoreEvent:
        if self._highscore_events:
            return self._highscore_events[-1]

    def get_loss_plot_points(self) -> list[tuple[int, float]]:
        return [(e.epoch, e.loss) for e in self._losses]

    def get_recent_loss_plot_points(self) -> list[tuple[int, float]]:
        return [(e.epoch, e.loss) for e in self._recent_losses]

    def get_scores_dist_plot_points(self) -> tuple[list[int], list[int]]:
        scores = sorted(self._scores_dist.keys())
        counts = [self._scores_dist[score] for score in scores]
        return scores, counts

    def get_recent_scores_dist_plot_points(
        self,
    ) -> tuple[list[int], list[int]]:
        scores = sorted(self._recent_scores_dist.keys())
        counts = [self._recent_scores_dist[score] for score in scores]
        return scores, counts

    def get_scores_dist_stats(self) -> tuple[float, float] | None:
        if not self._scores_dist_list:
            return None
        return mean(self._scores_dist_list), median(self._scores_dist_list)

    def get_recent_scores_dist_stats(self) -> tuple[float, float] | None:
        if not self._recent_scores_dist_list:
            return None
        return mean(self._recent_scores_dist_list), median(
            self._recent_scores_dist_list
        )

    def set_initial_epsilon(self, value: float) -> None:
        self._cur_epsilon = value



# ai_hydra/server/HydraServer.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

import argparse
import asyncio
import signal
from typing import Callable, Optional

from ai_hydra.constants.DHydra import (
    DHydraLog,
    DHydraLogDef,
    DHydraRouterDef,
    DHydraServerDef,
    DMethod,
    DModule,
)
from ai_hydra.utils.HydraLog import HydraLog
from ai_hydra.zmq.HydraMsg import HydraMsg
from ai_hydra.zmq.HydraServerMQ import HydraServerMQ


class HydraServer:
    """
    Abstract base class for HydraServer implementations.

    Provides common ZeroMQ-based server functionality that binds to a port
    and handles client requests using the REQ/REP pattern. Subclasses must
    implement application-specific message handling logic.
    """

    def __init__(
        self,
        address: str = "*",
        port: int = DHydraServerDef.PORT,
        router_address: str = DHydraRouterDef.HOSTNAME,
        router_port: int = DHydraRouterDef.PORT,
        identity: str = DModule.HYDRA_SERVER,
        log_level: DHydraLog = DHydraLogDef.DEFAULT_LOG_LEVEL,
    ):
        """
        Initialize the HydraServer with binding parameters.

        Args:
            address (str): The address to bind to (default: "*" for all
                interfaces)
            port (int): The port to bind to
            server_id (str): Identifier for logging purposes
        """

        self.address = address
        self.port = port
        self.router_address = router_address
        self.router_port = router_port
        self.identity = identity
        self.log_level = log_level

        self._methods: dict[str, Callable[[HydraMsg], object]] = {}

        # Messaging stub
        self.mq: Optional[HydraServerMQ] = None

        # Structured console logs
        self.log = HydraLog(
            client_id=self.identity, log_level=log_level, to_console=True
        )

        # Shutdown coordination
        self._stop_event: asyncio.Event | None = None
        self._main_task: asyncio.Task[None] | None = None

    def _install_signal_handlers(self, stop_event: asyncio.Event) -> None:
        """
        Prefer signal handlers when available (Unix). On Windows,
        add_signal_handler may not be implemented; Ctrl-C will still raise
        KeyboardInterrupt.
        """
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            return

        def _request_stop() -> None:
            if not stop_event.is_set():
                self.log.info("Shutdown requested (signal)")
                stop_event.set()

        for sig in (signal.SIGINT, signal.SIGTERM):
            try:
                loop.add_signal_handler(sig, _request_stop)
            except (NotImplementedError, RuntimeError):
                # Not supported on some platforms/event loops
                pass

    def loglevel(self, log_level: DHydraLog) -> None:
        """
        Initialize console logging for the server instance.
        """
        self.log = HydraLog(
            client_id=self.identity, log_level=log_level, to_console=True
        )

    async def quit(self):
        await self.mq.quit()

    async def _main_loop(self, stop_event: asyncio.Event) -> None:
        self.mq = HydraServerMQ(
            router_address=self.router_address,
            router_port=self.router_port,
            identity=self.identity,
            srv_methods=self._methods,
            log_level=self.log_level,
        )
        self.mq.start()

        try:
            while not stop_event.is_set():
                await asyncio.sleep(1)

        except asyncio.CancelledError:
            # Normal during shutdown
            raise

    async def ping(self, msg: HydraMsg) -> None:
        self.log.info(f"Received ping from {msg.sender}")
        reply_msg = HydraMsg(
            sender=DModule.HYDRA_SERVER,
            target=msg.sender,
            method=DMethod.PONG,
        )
        if self.mq is not None:
            await self.mq.send(reply_msg)
            self.log.info(f"Sent pong to {msg.sender}")

    async def run(self) -> None:
        """Entrypoint for running the server inside an event loop."""
        self._stop_event = asyncio.Event()
        stop_event = self._stop_event  # local alias for closure safety
        self._install_signal_handlers(stop_event)

        self.log.info("Initialized")
        self._main_task = asyncio.create_task(
            self._main_loop(stop_event), name="hydra-server-main"
        )

        try:
            await self._stop_event.wait()
        finally:
            await self._shutdown()

    async def _shutdown(self) -> None:
        """Graceful shutdown: cancel main task, close MQ, cancel stragglers."""
        self.log.info("Shutting down...")

        # Cancel main loop task if still running
        if self._main_task is not None and not self._main_task.done():
            self._main_task.cancel()
            try:
                await self._main_task
            except asyncio.CancelledError:
                pass

        # Stop/close MQ if available
        if self.mq is not None:
            stop = getattr(self.mq, "stop", None)
            if callable(stop):
                try:
                    maybe_awaitable = stop()
                    if asyncio.iscoroutine(maybe_awaitable):
                        await maybe_awaitable
                except Exception as e:
                    self.log.error(f"Error stopping HydraMQ: {e}")
            self.mq = None

        # Cancel any other pending tasks (keeps asyncio.run() happy)
        current = asyncio.current_task()
        pending = [
            t for t in asyncio.all_tasks() if t is not current and not t.done()
        ]
        for t in pending:
            t.cancel()
        if pending:
            await asyncio.gather(*pending, return_exceptions=True)

        self.log.info("Shutdown complete.")


async def amain() -> None:
    parser = argparse.ArgumentParser(description="Hydra ZeroMQ server")
    parser.add_argument("--address", default="*", help="Address to bind to")
    parser.add_argument(
        "--log_level",
        default=DHydraLogDef.DEFAULT_LOG_LEVEL,
        help="Log level [DEBUG|INFO|WARNING|ERROR|CRITICAL]",
    )
    parser.add_argument(
        "--port", type=int, default=DHydraServerDef.PORT, help="Server port"
    )
    parser.add_argument(
        "--router-address",
        default=DHydraRouterDef.HOSTNAME,
        help="HydraRouter hostname",
    )
    parser.add_argument(
        "--router-port",
        type=int,
        default=DHydraRouterDef.PORT,
        help="HydraRouter main port",
    )
    parser.add_argument(
        "--identity", default=DModule.HYDRA_SERVER, help="Server identity"
    )
    args = parser.parse_args()

    server = HydraServer(
        address=args.address,
        port=args.port,
        router_address=args.router_address,
        router_port=args.router_port,
        identity=args.identity,
        log_level=args.log_level,
    )

    await server.run()


def main() -> None:
    try:
        asyncio.run(amain())
    except KeyboardInterrupt:
        # If signal handlers aren't supported, Ctrl-C lands here.
        # asyncio.run will already have cancelled tasks; keep this quiet and
        # clean.
        pass


if __name__ == "__main__":
    main()



# ai_hydra/nnet/ReplayMemory.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0


from __future__ import annotations

import traceback
from collections import deque
from random import Random
from typing import Deque, TypeVar

from ai_hydra.constants.DReplayMemory import DMemory
from ai_hydra.constants.DHydra import DHydraLog
from ai_hydra.constants.DNNet import DRNN

from ai_hydra.nnet.Transition import Transition
from ai_hydra.utils.HydraLog import HydraLog

MAX_CHUNKS = DMemory.MAX_CHUNKS
MIN_CHUNKS = DMemory.MIN_CHUNKS

MAX_MEM_SIZE = DMemory.MAX_MEM_SIZE  # Max Linear transitions
MIN_FRAMES = DMemory.MIN_FRAMES  # Min transitions before returning samples
# Bucket all chunks at or above this chunk index into one final bucket.
# Example with 4:
#   0 -> first chunk in a game
#   1 -> second chunk
#   2 -> third chunk
#   3 -> fourth chunk
#   4 -> fifth and beyond
MAX_CHUNK_BUCKET = 19

T = TypeVar("T")


class ReplayMemory:
    def __init__(
        self,
        rng: Random,
        log_level: DHydraLog,
        seq_length: int | None = None,
    ):
        self.log = HydraLog(
            client_id="ReplayMemory",
            log_level=log_level,
            to_console=True,
        )
        self._rng = rng

        ### RNN model settings
        self._cur_game: list[Transition] = []
        self._chunks: list[tuple[int, list[Transition]]] = []
        self._seq_len = seq_length
        self._memory_cold = True
        self._memory_not_full = True
        self._memory_not_balanced = True

        # Global cap across all RNN buckets
        self._max_chunks = MAX_CHUNKS
        self._num_chunks = 0

        # Counter
        self._samples_delivered = 0

        # bucket_idx -> rolling list of chunks
        self._chunks_by_idx: dict[int, list[list[Transition]]] = {
            idx: [] for idx in range(MAX_CHUNK_BUCKET + 1)
        }

        if seq_length is None:
            raise ValueError("Sequence length must be set")
        self.log.info("Initialized for RNN model training")
        self.log.info(f"Setting sequence length to {seq_length}")
        self.log.info(
            f"Setting maximum number of stored sequence to {MAX_CHUNKS}"
        )
        self.log.info(
            f"Setting maximum chunk bucket index to {MAX_CHUNK_BUCKET}"
        )

    def append(self, t: Transition) -> None:
        """Add a transition into memory"""

        self._cur_game.append(t)
        if t.done:
            self._finalize_game()

    def _finalize_game(self) -> None:
        game = self._cur_game
        self._cur_game = []

        if len(game) < self._seq_len:
            return

        new_chunks = self._chunk_from_end(game)

        for bucket_idx, chunk in new_chunks:
            self._chunks_by_idx[bucket_idx].append(chunk)
            self._num_chunks += 1

        self._prune_if_needed()

    def _prune_if_needed(self) -> None:
        if self._num_chunks <= self._max_chunks:
            return

        if self._memory_not_full:
            self.log.info(
                f"Memory has been filled to capacity: {self._max_chunks} sequences"
            )
            self._memory_not_full = False

        overflow = self._num_chunks - self._max_chunks

        # Simple pruning policy:
        # prune oldest chunks from the most overrepresented non-empty bucket first
        for _ in range(overflow):
            bucket_to_prune = self._largest_bucket_idx()
            if bucket_to_prune is None:
                break
            del self._chunks_by_idx[bucket_to_prune][0]
            self._num_chunks -= 1

    def _log_chuck_distro(self):
        is_balanced = True
        log_str = "Sequence distribution: ["
        for idx in self._chunks_by_idx.keys():
            bucket_size = len(self._chunks_by_idx[idx])
            log_str += f"{bucket_size}|"
            if bucket_size != (MAX_CHUNKS // (MAX_CHUNK_BUCKET + 1)):
                is_balanced = False
        self.log.debug(log_str[:-1] + "]")
        if is_balanced:
            self._memory_not_balanced = False
            self.log.info("Memory buckets are balanced")

    def _largest_bucket_idx(self) -> int | None:
        largest_idx: int | None = None
        largest_size = 0
        for idx, chunks in self._chunks_by_idx.items():
            size = len(chunks)
            if size > largest_size:
                largest_size = size
                largest_idx = idx
        return largest_idx

    def _chunk_from_end(
        self,
        game: list[Transition],
    ) -> list[tuple[int, list[Transition]]]:
        """
        Split a completed game into fixed-length chunks aligned from the end.

        Returns:
            List of (bucket_idx, chunk)
        """
        n = len(game)
        rem = n % self._seq_len
        start = rem if rem != 0 else 0

        chunks: list[tuple[int, list[Transition]]] = []
        chunk_idx = 0

        for i in range(start, n, self._seq_len):
            chunk = game[i : i + self._seq_len]
            if len(chunk) == self._seq_len:
                bucket_idx = min(chunk_idx, MAX_CHUNK_BUCKET)
                chunks.append((bucket_idx, chunk))
                chunk_idx += 1

        return chunks

    def num_chunks(self) -> int:
        return self._num_chunks

    def chunk_bucket_counts(self) -> dict[int, int]:
        return {
            idx: len(chunks) for idx, chunks in self._chunks_by_idx.items()
        }

    def _warmed_bucket_indices(self) -> list[int]:
        warmed: list[int] = []
        for idx, chunks in self._chunks_by_idx.items():
            if len(chunks) > 0:
                warmed.append(idx)
        return warmed

    def sample_chunks(
        self,
        batch_size: int,
    ) -> list[list[Transition]] | None:
        """
        Sample chunks across warmed chunk-index buckets.

        Sampling policy:
          - require enough chunks in aggregate
          - divide batch approximately evenly across non-empty buckets
          - redistribute remainder to buckets with spare capacity
        """
        if self._num_chunks < max(batch_size, MIN_CHUNKS):
            return None

        warmed = self._warmed_bucket_indices()
        if not warmed:
            return None

        if self._memory_cold:
            self.log.debug("Memory has warmed up")
            self._memory_cold = False

        samples: list[list[Transition]] = []

        # First pass: even split across warmed buckets
        base = batch_size // len(warmed)
        remainder = batch_size % len(warmed)

        requested: dict[int, int] = {}
        for idx in warmed:
            requested[idx] = base

        # distribute remainder from later buckets backward so late-game buckets
        # get slight preference when possible
        warmed_desc = sorted(warmed, reverse=True)
        for i in range(remainder):
            requested[warmed_desc[i % len(warmed_desc)]] += 1

        # Sample what we can from each bucket
        shortage = 0
        for idx in warmed:
            bucket = self._chunks_by_idx[idx]
            want = requested[idx]
            take = min(want, len(bucket))
            if take > 0:
                samples.extend(self._rng.sample(bucket, take))
            shortage += want - take

        if shortage > 0:
            # Fallback pool from all unused capacity
            flat_pool: list[list[Transition]] = []
            already_selected_ids = {id(chunk) for chunk in samples}

            for idx in warmed:
                for chunk in self._chunks_by_idx[idx]:
                    if id(chunk) not in already_selected_ids:
                        flat_pool.append(chunk)

            if len(flat_pool) < shortage:
                return None

            samples.extend(self._rng.sample(flat_pool, shortage))

        self._samples_delivered += 1

        if self._memory_not_balanced and self._samples_delivered % 500 == 0:
            self._log_chuck_distro()

        return samples



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from ai_hydra.server.SnakeMgr import SnakeMgr
from ai_hydra.constants.DGame import DGameField

from ai_hydra.nnet.Policy.HydraPolicy import HydraPolicy

# from ai_hydra.nnet.UNUSED_RNNTrainer import RNNTrainer
from ai_hydra.nnet.LinearTrainer import LinearTrainer
from ai_hydra.nnet.ReplayMemory import ReplayMemory
from ai_hydra.nnet.Transition import Transition


@dataclass(frozen=True)
class EpisodeStats:
    episode_id: int
    steps: int
    score: int
    reward_total: int
    losses: int
    loss_avg: float


@dataclass(frozen=True)
class TrainRunStats:
    episodes: int
    avg_score: float
    avg_steps: float
    avg_reward_total: float
    avg_loss: float


class TrainMgr:
    """
    Training orchestrator.

    Owns:
      - policy (possibly epsilon-greedy wrapper)
      - trainer (torch + optimizer)
      - replay memory

    Uses:
      - SnakeMgr as the environment/session API

    Does NOT:
      - talk to MQ
      - render
    """

    def __init__(
        self,
        *,
        snake_mgr: SnakeMgr,
        policy: HydraPolicy,
        trainer,
        replay: ReplayMemory,
        client_id: str = "TrainMgr",
    ) -> None:
        self.snake_mgr = snake_mgr
        self.policy = policy
        self.trainer = trainer
        self.replay = replay
        self.client_id = client_id



# ai_hydra/client/plots/GameScorePlot.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from textual.app import ComposeResult, Widget
from textual.containers import Horizontal
from textual.widgets import Label
from textual_plot import HiResMode, LegendLocation, PlotWidget

from ai_hydra.constants.DHydraTui import DField, DLabel, DColor, DPlotDef
from ai_hydra.client.plots.PlotUtils import thin_series
from ai_hydra.utils.HydraMetrics import HydraMetrics


class LossPlot(Widget):

    def __init__(self, metrics: HydraMetrics, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.metrics = metrics

    def compose(self) -> ComposeResult:
        yield Horizontal(
            PlotWidget(id=DField.PLOT_LOSS),
            PlotWidget(id=DField.PLOT_RECENT_LOSS),
        )

    def plot_all(self):
        self.plot_loss()
        self.plot_recent_loss()

    def plot_loss(self):
        loss_events = self.metrics.get_loss_plot_points()
        if not loss_events:
            return

        episodes, losses = zip(*loss_events)
        if len(losses) > DPlotDef.MAX_LOSS_DATA_POINTS:
            episodes, losses = thin_series(
                x=episodes,
                y=losses,
                max_points=DPlotDef.MAX_LOSS_DATA_POINTS,
            )

        self._plot(tag=DField.ALL, episodes=episodes, losses=losses)

    def plot_recent_loss(self):
        loss_events = self.metrics.get_recent_loss_plot_points()
        if not loss_events:
            return

        episodes, losses = zip(*loss_events)
        self._plot(tag=DField.RECENT, episodes=episodes, losses=losses)

    def _plot(self, tag, episodes, losses):

        if tag == DField.ALL:
            plot = self.query_one(f"#{DField.PLOT_LOSS}", PlotWidget)
            ylabel = DLabel.LOSS
        elif tag == DField.RECENT:
            plot = self.query_one(f"#{DField.PLOT_RECENT_LOSS}", PlotWidget)
            ylabel = DLabel.RECENT_LOSS

        plot.clear()

        plot.plot(
            x=episodes,
            y=losses,
            line_style=DColor.GREEN,
            hires_mode=HiResMode.BRAILLE,
            label=DLabel.LOSS,
        )
        plot.set_ylimits()
        plot.set_xlabel(DLabel.EPISODES)
        plot.set_ylabel(ylabel)



# ai_hydra/nnet/Policy/RNNPolicy.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

import torch
from typing import Sequence

from ai_hydra.nnet.Policy.HydraPolicy import HydraPolicy
from ai_hydra.nnet.models.RNNModel import RNNModel


class RNNPolicy(HydraPolicy):
    def __init__(self, model: RNNModel, device: torch.device | None = None):
        self.model = model
        self.device = device or torch.device("cpu")
        self.model.eval()

    def select_action(self, state: Sequence[float]) -> int:
        with torch.no_grad():
            x = torch.tensor(
                state,
                dtype=torch.float32,
                device=self.device,
            )

            q = self.model(x)  # [3]
            return int(q.argmax().item())



# ai_hydra/client/HighScoresLog.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from textual.containers import Vertical
from textual.widgets import TabbedContent, Label, Log
from textual.app import ComposeResult, Widget

from ai_hydra.constants.DHydraTui import DField, DLabel, DStatus
from ai_hydra.utils.MetricEvent import HighscoreEvent


class HighScoresLog(Widget):

    def compose(self) -> ComposeResult:
        yield Vertical(
            Label(
                f"[b #3e99af]{DLabel.GAME:>7s}{DLabel.SCORE:>7s}{DLabel.TIME:>13s}[/]"
            ),
            Log(highlight=False, auto_scroll=True, id=DField.HIGHSCORES),
            classes=DField.HIGHSCORES_BOX,
        )

    def add_highscore(self, event: HighscoreEvent):
        self.query_one(f"#{DField.HIGHSCORES}", Log).write_line(
            f"{event.epoch:7d}{event.highscore:7d}{event.elapsed_time:>13s}"
        )



# ai_hydra/client/EventLog.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from textual.containers import Vertical
from textual.widgets import TabbedContent, Label, Log
from textual.app import ComposeResult, Widget

from ai_hydra.constants.DHydraTui import DField, DLabel
from ai_hydra.constants.DEvent import EVENT_MAP

from ai_hydra.utils.HydraMetrics import HydraMetrics


class EventLog(Widget):

    def __init__(self, metrics: HydraMetrics, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.metrics = metrics

    def compose(self) -> ComposeResult:
        yield Vertical(
            Label(
                f"[b #3e99af]   {DLabel.TYPE:<14s}{DLabel.TIME:>10s}  {DLabel.EVENT}[/]"
            ),
            Log(highlight=True, auto_scroll=True, id=DField.EVENT_LOG_LOG),
        )

    def add_event(self, ev_type: str, event: str):
        elap_time = self.metrics.get_elapsed_time()

        ev_type_str = EVENT_MAP[ev_type]

        self.query_one(f"#{DField.EVENT_LOG_LOG}", Log).write_line(
            f"{ev_type_str:<17s}{elap_time:>9s}  {event}"
        )



# ai_hydra/utils/HydraSnapshot.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0


from datetime import datetime

from ai_hydra.constants.DHydra import DHydra
from ai_hydra.constants.DHydraTui import DLabel, DField
from ai_hydra.constants.DNNet import (
    DNetDef,
    DLinear,
    DRNN,
    DNetField,
    MODEL_TYPE_TABLE,
)

from ai_hydra.utils.SimCfg import SimCfg
from ai_hydra.utils.HydraMetrics import HydraMetrics


class HydraSnapshot:

    def __init__(self, metrics: HydraMetrics):
        self.metrics = metrics

    def create_snapshot(self, snap_file, cfg: SimCfg):

        now = datetime.now()
        timestamp = now.strftime("%Y-%m-%d %H:%M:%S")

        # Model info
        model_type = cfg.get(DNetField.MODEL_TYPE)
        model_hidden_size = cfg.get(DNetField.HIDDEN_SIZE)
        model_input_size = DNetDef.INPUT_SIZE
        dropout_p = cfg.get(DNetField.DROPOUT_P)
        rnn_layers = cfg.get(DNetField.RNN_LAYERS)
        # Epsilon info
        initial_epsilon = cfg.get(DNetField.INITIAL_EPSILON)
        min_epsilon = cfg.get(DNetField.MIN_EPSILON)
        epsilon_delay = cfg.get(DNetField.EPSILON_DECAY)
        # Training info
        learning_rate = cfg.get(DNetField.LEARNING_RATE)
        gamma = cfg.get(DNetField.GAMMA)
        batch_size = cfg.get(DNetField.BATCH_SIZE)
        seq_length = cfg.get(DNetField.SEQ_LENGTH)
        rnn_tau = cfg.get(DNetField.RNN_TAU)

        # Metrics data
        elapsed_time = self.metrics.get_elapsed_time()
        cur_epoch = self.metrics.get_cur_epoch()

        with open(snap_file, "w") as f:
            f.write(
                "📸 AI Hydra - Snapshot\n"
                "══════════════════════\n"
                f"Timestamp: {timestamp}\n"
                f"Simulation Run Time: {elapsed_time}\n"
                f"Current Episode Number: {cur_epoch}\n"
                f"AI Hydra Version: v{DHydra.VERSION}\n"
                f"Random Seed: {DHydra.RANDOM_SEED}\n\n"
            )

            f.write(
                "🎯 Epsilon Greedy\n"
                "═════════════════\n"
                f"Initial Epsilon: {initial_epsilon}\n"
                f"Minimum Epsilon: {min_epsilon}\n"
                f"Epsilon Decay Rate: {epsilon_delay}\n\n"
            )
            if model_type == DField.LINEAR:
                f.write(
                    "🧠 Linear Model\n"
                    "═══════════════\n"
                    f"Input Size: {model_input_size}\n"
                    f"Hidden Size: {model_hidden_size}\n"
                    f"Dropout Layer P-Value: {dropout_p}\n\n"
                    "🧙 Training\n"
                    "═══════════\n"
                    f"Learning Rate: {learning_rate}\n"
                    f"Discount/Gamma: {gamma}\n"
                    f"Batch Size: {batch_size}\n\n"
                )
            elif model_type == DField.RNN:
                f.write(
                    "🧠 RNN Model\n"
                    "════════════\n"
                    f"Input Size: {model_input_size}\n"
                    f"Hidden Size: {model_hidden_size}\n"
                    f"Dropout Layer P-Value: {dropout_p}\n"
                    f"RNN Layers: {rnn_layers}\n"
                    f"RNN Tau: {rnn_tau}\n\n"
                    "🧙 Training\n"
                    "═══════════\n"
                    f"Learning Rate: {learning_rate}\n"
                    f"Discount/Gamma: {gamma}\n"
                    f"Batch Size: {batch_size}\n"
                    f"Sequence Length: {seq_length}\n"
                    f"RNN Tau: {rnn_tau}\n\n"
                )
            f.write(
                "🏆 Highscore Events\n"
                "═══════════════════\n"
                f"{'Episode':8s}{'Highscore':<10s}{'Time':>11s}{'Epsilon':>8s}\n"
                "═══════ ═════════ ═══════════ ═══════\n"
            )
            rows = self.metrics.get_highscore_snapshot_rows()
            for row in rows:
                epoch, highscore, epsilon, ev_time = row
                cur_ep = str(round(float(epsilon), 4))

                f.write(
                    f"{str(epoch):>7s}{str(highscore):>10s}{ev_time:>12s}{cur_ep:>8s}\n"
                )
